"""
Connection Object that uses Python Requests
"""
from arcgis.auth.tools import LazyLoader
from typing import Union
from arcgis.auth.tools._util import check_module_exists

try:
    arcpy = LazyLoader("arcpy", strict=True)

    HASARCPY = True
except ImportError:
    HASARCPY = False
except:
    HASARCPY = False

try:
    requests_gssapi = LazyLoader("requests_gssapi", strict=True)
    HAS_GSSAPI = True
except:
    HAS_GSSAPI = False

import sys

if sys.platform == "win32" and check_module_exists("certifi_win32"):
    try:
        import certifi_win32

        certifi_win32.wincerts.where()

        if certifi_win32.wincerts.verify_combined_pem() == False:
            certifi_win32.generate_pem()

    except ImportError:
        pass
elif check_module_exists("truststore"):  # pragma: no cover
    try:
        import truststore

        truststore.inject_into_ssl()
    except ImportError as ie:
        pass
    except Exception as e:
        pass

import os
import copy
import json
import uuid
import datetime
import mimetypes
import logging
import warnings
import tempfile
from functools import lru_cache
from urllib.request import urlparse
import requests
from urllib3 import exceptions as _exceptions

# from requests import Session
from requests_toolbelt.downloadutils import stream
from json import JSONDecodeError
from ._helpers import _filename_from_headers, _filename_from_url
from ._authguess import GuessAuth
from arcgis._impl.common._utils import _date_handler
from arcgis._impl.common._mixins import PropertyMap
from arcgis._impl.common._isd import InsensitiveDict
from arcgis.auth import EsriSession
from arcgis.auth import (
    EsriBuiltInAuth,
    EsriAPIKeyAuth,
    EsriGenTokenAuth,
    ArcGISProAuth,
    EsriOAuth2Auth,
    EsriUserTokenAuth,
)
from arcgis.auth._auth._notebook import EsriNotebookAuth

try:
    from arcgis.auth import EsriWindowsAuth

    HAS_SSPI = True
except ImportError:
    HAS_SSPI = False

try:
    from arcgis.auth import EsriKerberosAuth

    HAS_KERBEROS = True
except ImportError:
    HAS_KERBEROS = False

from arcgis.auth import EsriBasicAuth

__version__ = "2.2.0.1"

_DEFAULT_TOKEN = uuid.uuid4()
_log = logging.getLogger(__name__)


class Connection(object):
    """
    Universal Connection Object
    """

    _timeout = None
    _refresh_token = None
    _token = None
    _token_url = None
    _create_time = None
    _session = None
    _header = None
    _username = None
    _password = None
    _proxy_url = None
    _proxy_port = None
    _proxy_password = None
    _proxy_username = None
    _verify_cert = None
    _baseurl = None
    baseurl = None
    _referer = None
    _expiration = None
    _auth = None
    _product = None
    _custom_auth = None
    _custom_adapter = None
    legacy = None
    _server_log = None

    # ----------------------------------------------------------------------
    def __init__(
        self,
        baseurl: str = None,
        username: str = None,
        password: str = None,
        **kwargs,
    ):
        """initializer

        Optional Kwargs

        client_id=None
        client_secret=None
        expiration=60
        cert_file
        key_file
        proxy_url
        proxy_port
        proxy : Dict - Dict
        verify_cert
        referer
        all_ssl = True
        portal_connection = None
        product = "UNKNOWN", "PORTAL", "SERVER", "AGOL", "GEOEVENT"  # SHOULD BE SET BY GIS object, or Server Object, etc...
        token= None If token, _AUTH = BUILTIN
        token_url
        AUTH keys = HOME, BUILTIN, PRO, ANON, PKI, HANDLER, UNKNOWN (Internal)
        custom_auth = Requests authencation handler
        trust_env = T/F if to ignore netrc files
        legacy boolean. If True the token will be appended to the URL for GET and in the FORM POST.
        timeout:int=600
        use_gen_token = boolean - Uses the GenTokenAuth over EsriBuiltInAuth
        security_kwargs = dict - a set of optional arguments for Kerberos Auth
        """
        from arcgis.gis import GIS

        self._ags_file = kwargs.pop("ags_file", None)
        self._security_kwargs = kwargs.pop("security_kwargs", {})
        self._use_gen_token = kwargs.pop("use_gen_token", False)
        self._is_hosted_nb_home = kwargs.pop("is_hosted_nb_home", False)
        self._proxy = kwargs.pop("proxy", None)
        self._timeout = kwargs.pop("timeout", 600)
        self._all_ssl = kwargs.pop("all_ssl", True)
        self.trust_env = kwargs.pop("trust_env", None)
        self._custom_adapter = kwargs.pop("custom_adapter", None)
        self.legacy = kwargs.pop("legacy", False)
        if baseurl:
            while baseurl.endswith("/"):
                baseurl = baseurl[:-1]
        self._custom_auth = kwargs.pop("custom_auth", None)
        self._product = kwargs.pop("product", "UNKNOWN")
        if baseurl is None:
            self._product = "AGOL"
            baseurl = "https://www.arcgis.com"

        self._baseurl = baseurl
        self._username = username
        self._password = password

        self._expiration = kwargs.pop("expiration", 60) or 60
        if self._expiration < 60:
            _log.warning("Expiration is less than 60 seconds, changing value to 60")
            self._expiration = 60
        self._portal_connection = kwargs.pop(
            "portal_connection", None
        )  # For Federated Objects (Portal Connection)
        if self._ags_file and os.path.isfile(self._ags_file):
            res = arcpy.gp.getStandaloneServerToken(self._ags_file)
            self._referer = res.pop("referer", "http")
            self._baseurl = res.pop("serverUrl", None)
            baseurl = self._baseurl
        elif isinstance(self._portal_connection, GIS):
            self._portal_connection = self._portal_connection._con

        if (
            (self._referer or self._referer is None)
            and self._portal_connection
            and str(self._portal_connection._auth).lower() == "home"
        ):
            self._referer = None
        elif baseurl.lower() == "pro":
            self._auth = "PRO"
            try:
                self._referer = arcpy.GetSigninToken().pop("referer", "http")
            except:
                self._referer = kwargs.pop("referer", "http")
        else:
            self._referer = kwargs.pop("referer", "http")

        self._verify_cert = kwargs.pop("verify_cert", False)  # True)
        if self._verify_cert == False:
            warnings.simplefilter("ignore", _exceptions.InsecureRequestWarning)

        self._cert_file = kwargs.pop("cert_file", None)
        self._key_file = kwargs.pop("key_file", None)
        if "proxy_host" in kwargs:
            self._proxy_url = kwargs.pop("proxy_host", None)
        else:
            self._proxy_url = kwargs.pop("proxy_url", None)
        self._proxy_port = kwargs.pop("proxy_port", None)
        self._proxy_username = kwargs.pop("proxy_username", None)
        self._proxy_password = kwargs.pop("proxy_password", None)
        self._header = {"User-Agent": "Geosaurus/%s" % __version__}
        self._client_id = kwargs.pop("client_id", None)
        self._client_secret = kwargs.pop("client_secret", None)
        self._token_url = kwargs.pop("token_url", None)
        if str(baseurl).lower() == "pro":
            self._auth = "PRO"
            auth_check = [""]
        elif str(baseurl).lower() != "pro":
            auth_check = [""]
        if self._is_hosted_nb_home:
            auth_check = [""]
        elif self._key_file is None and self._cert_file is None:
            auth_check = self._auth_check(baseurl, proxies=self._assemble_proxy())
        else:
            auth_check = [""]
        if self._is_hosted_nb_home:
            self._auth = "HOME"  # NB AUTH
            self._token = kwargs.pop("token", None)
            self._expiration = 10080
            self._referer = ""
        elif self._ags_file:
            self._auth = "AGS_AUTH"  # AGS AUTH
        elif "token" in kwargs and kwargs["token"]:
            self._auth = "USER_TOKEN"
            self._token = kwargs.pop("token", None)
        elif "api_key" in kwargs and kwargs["api_key"]:
            self._auth = "API_KEY"
            self._api_key = kwargs.pop("api_key", None)
        elif (
            self._key_file is None
            and self._key_file is None
            and username is None
            and password is None
            and self._portal_connection is None
            and self._client_id is None
            and str(baseurl).lower() != "pro"
            and any(
                [
                    a in auth_check
                    for a in [
                        "Negotiate",
                        "NTLM",
                        "Negotiate, NTLM",
                        "Basic",
                        "basic",
                    ]
                ]
            )
            == False
        ):
            self._auth = "ANON"
        elif self._client_id:
            self._auth = "OAUTH"
        elif (not username is None and not password is None) and (
            "Negotiate" in auth_check or "Negotiate, NTLM" in auth_check
        ):
            self._auth = "KERBEROS"
        elif (username is None and password is None) and (
            "Negotiate" in auth_check or "Negotiate, NTLM" in auth_check
        ):
            self._auth = "KERBEROS"
        elif (not username is None and not password is None) and len(
            username.split("\\")
        ) > 1:
            self._auth = "IWA"
        elif (
            not username is None
            and not password is None
            and any([ac.lower().find("basic") > -1 for ac in auth_check])
        ):
            self._auth = "BASIC_REALM"
        elif (
            not username is None
            and not password is None
            and any([ac.lower().find("ntlm") > -1 for ac in auth_check])
        ):
            self._auth = "NTLM"
        elif (not username is None and not password is None) or (
            self._portal_connection and self._portal_connection._auth == "BUILTIN"
        ):
            self._auth = "BUILTIN"
        elif (not username is None and not password is None) or (
            self._portal_connection and self._portal_connection._auth == "BASIC_REALM"
        ):
            self._auth = "BASIC_REALM"
        elif (not username is None and not password is None) or (
            self._portal_connection and self._portal_connection._auth == "NTLM"
        ):
            self._auth = "NTLM"
        elif (
            (username and password)
            and self._client_id is None
            and str(baseurl).lower() != "pro"
        ):
            self._auth = "BUILTIN"
        elif self._portal_connection:
            self._auth = "BUILTIN"
        elif baseurl.lower() == "pro":
            self._auth = "PRO"
            portal_url = arcpy.GetActivePortalURL()
            if portal_url.lower().find("/sharing/rest") == -1:
                if arcpy.GetActivePortalURL().endswith("/"):
                    self._baseurl = arcpy.GetActivePortalURL() + "sharing/rest"
                else:
                    self._baseurl = arcpy.GetActivePortalURL() + "/sharing/rest"
            else:
                self._baseurl = arcpy.GetActivePortalURL()
        elif self._cert_file or (self._cert_file and self._key_file):
            self._auth = "PKI"

        if self._portal_connection and self._portal_connection._auth in [
            "BASIC_REALM",
            "IWA",
            "NTLM",
            "KERBEROS",
            "PKI",
        ]:
            self._session = self._portal_connection._session
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self._create_session()

        #  Product Info
        if self._client_id:
            self._product = "PORTAL"
        elif self._is_hosted_nb_home:
            self._product = "NOTEBOOK_SERVER"
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self._product = self._check_product()
        self._baseurl = self._validate_url(self._baseurl)
        self.baseurl = self._baseurl

    # ----------------------------------------------------------------------
    @lru_cache(maxsize=10)
    def _parsed(self, url: str):
        return urlparse(url)

    # ----------------------------------------------------------------------
    def _auth_check(self, url: str, proxies: dict = None) -> list:
        import requests

        if str(url).lower() == "pro":
            return [""]
        if self._cert_file and self._key_file:
            cert = (self._cert_file, self._key_file)

        elif self._cert_file and self._password:
            from arcgis.gis._impl._con._cert import pfx_to_pem

            self._key_file, self._cert_file = pfx_to_pem(
                pfx_path=self._cert_file, pfx_password=self._password
            )
            cert = (self._cert_file, self._key_file)
        elif self._cert_file:
            cert = self._cert_file
        else:
            cert = None
        with requests.Session() as s:
            s.cert = cert
            s.verify = self._verify_cert
            s.trust_env = True
            if self._custom_adapter:
                for k, v in self._custom_adapter.items():
                    s.mount(k, v)
            if self._custom_auth:
                s.auth = self._custom_auth
            parsed = self._parsed(url)
            root = (
                rf"{parsed.scheme}://{parsed.netloc}/{parsed.path[1:].split(r'/')[0]}"
            )
            params = {"f": "json"}
            results = []
            for pt in [
                "/info",
                "/rest/info",
                "/sharing/rest/info",
                "/rest/services",
            ]:
                try:
                    www_auth = s.get(
                        root + pt,
                        params=params,
                        verify=self._verify_cert,
                        proxies=proxies,
                    ).headers.get("www-authenticate", "")
                    results.append(www_auth)
                except:
                    results.append("")
        return list(set(results))

    # ----------------------------------------------------------------------
    def _validate_url(self, url: str) -> str:
        """ensures the base url has the /sharing/rest"""
        if url.lower().find("arcgis.com") > -1:
            self._product = "AGOL"
        if self._product in ["AGO", "AGOL", "PORTAL"]:
            if not url[-1] == "/":
                url += "/"
            if url.lower().find("www.arcgis.com") > -1:
                urlscheme = urlparse(url).scheme
                return "{scheme}://www.arcgis.com/sharing/rest/".format(
                    scheme=urlscheme
                )
            elif url.lower().endswith("sharing/"):
                return url + "rest/"
            elif url.lower().endswith("sharing/rest/"):
                return url
            else:
                return url + "sharing/rest/"
        elif self._product in ["SERVER", "FEDERATED_SERVER", "FEDERATED"]:
            parsed = urlparse(url)
            path = parsed.path
            if str(path).startswith("/"):
                path = path[1:]
            url = "%s://%s/%s/rest/services/" % (
                parsed.scheme,
                parsed.netloc,
                path.split("/")[0],
            )
        return url

    # ----------------------------------------------------------------------
    def _assemble_proxy(self) -> dict:
        if self._proxy and isinstance(self._proxy, dict):
            return self._proxy
        elif self._proxy_port and self._proxy_url:
            url = "%s:%s" % (self._proxy_url, self._proxy_port)
            if self._proxy_password and self._proxy_username:
                proxies = {
                    "http": "http://%s:%s@%s"
                    % (self._proxy_username, self._proxy_password, url),
                    "https": "https://%s:%s@%s"
                    % (self._proxy_username, self._proxy_password, url),
                }
            else:
                proxies = {
                    "http": "http://%s" % url,
                    "https": "https://%s" % url,
                }
            return proxies
        return

    # ----------------------------------------------------------------------
    def _create_session(self) -> requests.Session:
        if self._proxy and isinstance(self._proxy, dict):
            proxies = self._proxy
            self._proxy = proxies
        elif self._proxy_port and self._proxy_url:
            url = "%s:%s" % (self._proxy_url, self._proxy_port)
            if self._proxy_password and self._proxy_username:
                proxies = {
                    "http": "http://%s:%s@%s"
                    % (self._proxy_username, self._proxy_password, url),
                    "https": "https://%s:%s@%s"
                    % (self._proxy_username, self._proxy_password, url),
                }
            else:
                proxies = {
                    "http": "http://%s" % url,
                    "https": "https://%s" % url,
                }
            self._proxy = proxies
        else:
            proxies = None

        if self._cert_file and self._key_file:
            cert = (self._cert_file, self._key_file)

        elif self._cert_file and self._password:
            from arcgis.gis._impl._con._cert import pfx_to_pem

            self._key_file, self._cert_file = pfx_to_pem(
                pfx_path=self._cert_file, pfx_password=self._password
            )
            cert = (self._cert_file, self._key_file)
        elif self._cert_file:
            cert = self._cert_file
        else:
            cert = None

        self._session = EsriSession(
            cert=cert,
            verify_cert=self._verify_cert,
            proxies=proxies,
            retries=5,
        )
        self._session.verify = self._verify_cert
        self._session.stream = True
        self._session.trust_env = self.trust_env
        self._session.headers.update(self._header)
        self._session.proxies = proxies

        from urllib3.util import Retry
        from urllib3 import __version__ as __urllib3_version__

        if self._custom_adapter:
            for k, v in self._custom_adapter.items():
                self._session.mount(k, v)

        if self._referer is None and (
            self._portal_connection
            and str(self._portal_connection._auth).lower() == "home"
        ):
            self._referer = "http"
            self._session.headers.pop("Referer", None)
            self._session.headers["Referer"] = json.dumps("")
        elif (
            self._portal_connection
            and str(self._portal_connection._auth).lower() == "home"
        ):
            self._referer = "http"
            self._session.headers.pop("Referer", None)
            self._session.headers["Referer"] = json.dumps("")
        elif self._referer is None:
            self._referer = "http"
            self._session.headers.update({"Referer": self._referer})
        else:
            self._session.headers.update({"Referer": self._referer})
        if self._custom_auth:
            self._session.auth = self._custom_auth
            self._auth = "CUSTOM"
        elif self._auth.lower() == "home":
            from arcgis.auth._auth._notebook import EsriNotebookAuth

            self._session.verify = False
            self._session.auth = EsriNotebookAuth(
                token=self._token,
                referer=self._referer,
                auth=GuessAuth(username=None, password=None),
            )
        elif self._auth.lower() == "ags_auth":
            from arcgis.auth._auth import ArcGISServerAuth

            self._session.auth = ArcGISServerAuth(
                legacy=self.legacy,
                ags_file=self._ags_file,
            )
        elif self._auth.lower() == "oauth":
            self._session.auth = EsriOAuth2Auth(
                base_url=self._baseurl,
                client_id=self._client_id,
                client_secret=self._client_secret,
                username=self._username,
                password=self._password,
                referer=self._referer,
                expiration=self._expiration,
                verify_cert=self._verify_cert,
                proxies=proxies,
            )
        elif self._auth.lower() == "builtin":
            if self._check_product() == "SERVER":
                pauth = None
                if self._portal_connection:
                    pauth = self._portal_connection._con._auth
                if self._token_url is None:
                    self._check_product()
                self._session.auth = EsriGenTokenAuth(
                    token_url=self._token_url,
                    referer=self._referer,
                    username=self._username,
                    password=self._password,
                    portal_auth=pauth,
                    time_out=self._timeout,
                    verify_cert=self._verify_cert,
                    legacy=self.legacy,
                    proxies=proxies,
                )
            else:
                if self._use_gen_token:
                    if self._token_url is None:
                        self._check_product()
                    self._session.auth = EsriGenTokenAuth(
                        token_url=self._token_url,
                        referer=self._referer,
                        username=self._username,
                        password=self._password,
                        portal_auth=None,
                        time_out=1440,
                        verify_cert=self._verify_cert,
                        legacy=self.legacy,
                        proxies=proxies,
                    )
                else:
                    self._session.auth = EsriBuiltInAuth(
                        url=self._baseurl,
                        username=self._username,
                        password=self._password,
                        expiration=self._timeout,
                        legacy=False,
                        verify_cert=self._verify_cert,
                        referer=self._referer,
                        proxies=proxies,
                    )
        elif self._auth.lower() == "user_token":
            self._session.auth = EsriUserTokenAuth(
                token=self._token,
                referer=self._referer,
                verify_cert=self._verify_cert,
            )
        elif self._auth.lower() == "api_key":
            from arcgis.auth._auth._apikey import EsriAPIKeyAuth

            self._session.auth = EsriAPIKeyAuth(
                api_key=self._api_key,
                referer=self._referer,
                verify_cert=self._verify_cert,
            )
            self._token = self._api_key
        elif self._auth.lower() == "basic_realm":
            self._session.auth = EsriBasicAuth(
                username=self._username,
                password=self._password,
                referer=self._referer,
                verify_cert=self._verify_cert,
            )
        elif self._auth.lower() in ["kerberos"] and HAS_KERBEROS:
            if self._security_kwargs:
                self._session.auth = EsriKerberosAuth(
                    proxies=self._proxy,
                    username=self._username,
                    password=self._password,
                    verify_cert=self._verify_cert,
                    legacy=False,
                    **self._security_kwargs,
                )
            elif HAS_GSSAPI:
                self._session.auth = EsriWindowsAuth(
                    username=self._username,
                    password=self._password,
                    verify_cert=self._verify_cert,
                    legacy=False,
                    proxies=self._proxy,
                )
            else:
                self._session.auth = EsriKerberosAuth(
                    proxies=self._proxy,
                    username=self._username,
                    password=self._password,
                    verify_cert=self._verify_cert,
                    legacy=False,
                    **self._security_kwargs,
                )
        elif self._username and self._password and self._auth.lower() != "iwa":
            self._session.auth = GuessAuth(
                username=self._username, password=self._password
            )
        elif self._auth.lower() in ["iwa", "ntlm"] and HAS_SSPI:
            self._session.auth = EsriWindowsAuth(
                username=self._username,
                password=self._password,
                verify_cert=self._verify_cert,
                legacy=False,
                proxies=self._proxy,
            )
        elif self._auth.lower() == "pro":
            self._session.auth = (
                GuessAuth(None, None, legacy=False) + ArcGISProAuth()
            )  # GuessAuth(None, None, legacy=False)
        elif not self._cert_file and not self._key_file:
            # else:

            if HAS_SSPI:
                try:
                    self._session.auth = EsriWindowsAuth(
                        verify_cert=self._verify_cert,
                        legacy=False,
                        proxies=self._proxy,
                    )
                except:
                    ...
            elif HAS_KERBEROS:
                try:
                    self._session.auth = EsriKerberosAuth(
                        verify_cert=self._verify_cert,
                        legacy=False,
                        proxies=self._proxy,
                    )
                except:
                    ...

    # ----------------------------------------------------------------------
    def get(
        self, path: str, params: dict = None, **kwargs
    ) -> Union[dict, requests.Response]:
        """

        sends a GET request.

        ===========================   =====================================================
        **optional keywords**         **description**
        ---------------------------   -----------------------------------------------------
        path                          required string. URI resource to access.
        ===========================   =====================================================

        ===========================   =====================================================
        **optional keywords**         **description**
        ---------------------------   -----------------------------------------------------
        params                        optional dictionary.  This is the data payload to the
                                      web resource
        ---------------------------   -----------------------------------------------------
        out_folder                    optional string. Save path on disk.  The default is
                                      the user's temp directory.
        ---------------------------   -----------------------------------------------------
        file_name                     optional string. Name of the file to save. If not
                                      provided, name will try to be parsed from the response.
        ---------------------------   -----------------------------------------------------
        try_json                      optional boolean.  If True, the marker f='json' will
                                      be appended to the parameters.
        ---------------------------   -----------------------------------------------------
        add_token                     optional boolean. Default True, if true, the call will
                                      add a token to any token based security.
        ---------------------------   -----------------------------------------------------
        json_encode                   optional Boolean.  When False, the JSON values will not be encoded.
        ---------------------------   -----------------------------------------------------
        ignore_error_key              optional Boolean. The default is False. If true, JSON will be returned and no exception is raised when 'error' is present in the response
        ---------------------------   -----------------------------------------------------
        return_raw_response           optional Boolean. Returns the requests' Response object.
        ---------------------------   -----------------------------------------------------
        allow_redirects               Optional Boolean. Does not allow the call to follow the 3xx status codes. The default is True
        ---------------------------   -----------------------------------------------------
        drop_auth                     Optional Boolean. Drop this when auth handlers cause issues. The default is False
        ===========================   =====================================================
        """
        drop_auth = kwargs.pop("drop_auth", False)
        ignore_error_key = kwargs.pop("ignore_error_key", False)
        allow_redirects = kwargs.pop("allow_redirects", True)
        return_raw_response = kwargs.pop("return_raw_response", False)
        json_encode = kwargs.pop("json_encode", True)
        add_headers = kwargs.pop("add_headers", {})
        if self._baseurl.endswith("/") == False:
            self._baseurl += "/"
        url = path
        if url.find("://") == -1:
            if url.startswith("/") == False and self._baseurl.endswith("/") == False:
                url = "/" + url
            url = self._baseurl + url
        if kwargs.pop("ssl", False) or self._all_ssl:
            url = url.replace("http://", "https://")
        if params is None:
            params = {}
        if self._session is None:
            self._create_session()

        try_json = kwargs.pop("try_json", True)
        if try_json and isinstance(params, dict):
            params["f"] = "json"
        if params == {}:
            params = None

        if "out_folder" in kwargs:
            out_path = kwargs.pop("out_folder", tempfile.gettempdir())
        else:
            out_path = kwargs.pop("out_path", tempfile.gettempdir())
        file_name = kwargs.pop("file_name", None)
        if isinstance(params, dict):
            if params and json_encode:
                for k, v in copy.copy(params).items():
                    if isinstance(v, (tuple, dict, list, bool)):
                        params[k] = json.dumps(v, default=_date_handler)
                    elif isinstance(v, PropertyMap):
                        params[k] = json.dumps(dict(v), default=_date_handler)
                    elif isinstance(v, InsensitiveDict):
                        params[k] = v.json
        if add_headers:
            original_headers = copy.deepcopy(self._session.headers)
            self._session.headers.update(add_headers)
        try:
            if self._cert_file:
                cert = (self._cert_file, self._key_file)
            else:
                cert = None
            if self._session.auth and drop_auth:
                auth = self._session.auth
                self._session.auth = None
                resp = self._session.get(
                    url=url,
                    params=params,
                    cert=cert,
                    verify=self._verify_cert,
                    allow_redirects=allow_redirects,
                )
                self._session.auth = auth
            else:
                resp = self._session.get(
                    url=url,
                    params=params,
                    cert=cert,
                    verify=self._verify_cert,
                    allow_redirects=allow_redirects,
                )
        except requests.exceptions.SSLError as err:
            raise requests.exceptions.SSLError(
                "Please set verify_cert=False due to encountered SSL error: %s" % err
            )
        except requests.exceptions.InvalidURL as errIU:
            raise requests.exceptions.SSLError("Invalid URL provided: %s" % errIU)
        except requests.exceptions.ConnectionError as errCE:
            raise requests.exceptions.ConnectionError(
                "A connection error has occurred: %s" % errCE
            )
        except requests.exceptions.InvalidHeader as errIH:
            raise requests.exceptions.InvalidHeader(
                "A invalid header was provided: %s" % errIH
            )
        except requests.exceptions.HTTPError as errh:
            raise requests.exceptions.HTTPError("Http Error: %s" % errh)
        except requests.exceptions.MissingSchema as errMS:
            raise requests.exceptions.MissingSchema(
                "URL scheme must be provided: %s" % errMS
            )
        except requests.exceptions.RequestException as errRE:
            raise requests.exceptions.RequestException(
                "A general exception was raised: %s" % errRE
            )
        except Exception as e:
            raise Exception("A general error occurred: %s" % e)
        except:
            import traceback

            raise Exception("An unknown error occurred: %s" % traceback.format_exc())
        if add_headers:
            self._session.headers.clear()
            self._session.headers.update(original_headers)
        if return_raw_response:
            return resp
        return self._handle_response(
            resp,
            file_name,
            out_path,
            try_json,
            force_bytes=kwargs.pop("force_bytes", False),
            ignore_error_key=ignore_error_key,
        )

    # ----------------------------------------------------------------------
    def _handle_response(
        self,
        resp: requests.Response,
        file_name: str,
        out_path: str,
        try_json: bool,
        force_bytes: bool = False,
        ignore_error_key: bool = False,
    ) -> dict:
        """
        handles the request responses

        ===========================   =====================================================
        **optional keywords**         **description**
        ---------------------------   -----------------------------------------------------
        resp                          required resquests.Rsponse object.
        ---------------------------   -----------------------------------------------------
        file_name                     required string.  Name of the output file if needed.
        ---------------------------   -----------------------------------------------------
        out_path                      required string. Name of the save folder.
        ---------------------------   -----------------------------------------------------
        try_json                      required boolean. Determines if the response should
                                      be returned as a dictionary or the native format
                                      send back by a server
        ---------------------------   -----------------------------------------------------
        force_bytes                   Optional Boolean. If True, the results returns as bytes
                                      instead of a file path.
        ===========================   =====================================================

        returns: string, dictionary, or bytes depending on the response.

        """
        if "Set-Cookie" in resp.headers and (
            str(self._auth).lower() in ["anon"]
            or (
                str(self._username).find("\\") > -1
                and str(self._auth).lower() in ["builtin"]
            )
        ):
            self._auth = "IWA"

        data = None
        url = resp.url
        if out_path and os.path.isdir(out_path) == False:
            os.makedirs(out_path)
        if out_path is None:
            out_path = tempfile.gettempdir()
        if (
            file_name is None
            and "Content-Type" in resp.headers
            and (
                resp.headers["Content-Type"].lower().find("json") == -1
                and resp.headers["Content-Type"].lower().find("text") == -1
            )
        ):
            file_name = (
                _filename_from_url(url) or _filename_from_headers(resp.headers) or None
            )
        elif file_name is None and "Content-Disposition" in resp.headers:
            file_name = (
                _filename_from_url(url) or _filename_from_headers(resp.headers) or None
            )
        if force_bytes:
            try:
                return bytes(resp.content)
            except:
                return resp.content
        if file_name is not None:
            file_name = os.path.join(out_path, file_name)
            if os.path.isfile(file_name):
                os.remove(file_name)
            stream_size = 512 * 2
            if "Content-Length" in resp.headers:
                max_length = int(resp.headers["Content-Length"])
                if max_length > stream_size * 2 and max_length < 1024 * 1024:
                    stream_size = 1024 * 2
                elif max_length > 5 * (1024 * 1024) and max_length < 10 * (1024 * 1024):
                    stream_size = 5 * (1024 * 1024)  # 5 mb
                elif max_length >= 10 * (1024 * 1024):
                    stream_size = 10 * (1024 * 1024)  # 10 mb
                elif max_length > (1024 * 1024):
                    stream_size = 1024 * 1024  # 1 mb
                else:
                    stream_size = 512 * 2

            fp = stream.stream_response_to_file(
                response=resp, path=file_name, chunksize=stream_size
            )
            return fp

        if try_json:
            if (
                "Content-Length" in resp.headers
                and int(resp.headers.get("Content-Length")) == 0
            ):
                data = {}
            elif (
                "Transfer-Encoding" in resp.headers
                and resp.headers["Transfer-Encoding"].lower() == "chunked"
            ):
                data = None
                for it in resp.iter_lines(
                    chunk_size=None, decode_unicode=True, delimiter=None
                ):
                    if data is None:
                        data = it
                    else:
                        data += it
                data = json.loads(data)
                if "error" in data and ignore_error_key == False:
                    raise Exception(data["error"])
            else:
                try:
                    data = resp.json()
                except JSONDecodeError:
                    if resp.text:
                        raise Exception(resp.text)
                    else:
                        raise
            # if 'error' in data:
            # raise Exception(data['error'])
            # return data
            # else:
            # return resp.text
            if "error" in data and ignore_error_key == False:
                if "messages" in data:
                    return data
                errorcode = data["error"]["code"] if "code" in data["error"] else 0
                self._handle_json_error(data["error"], errorcode)
            return data
        else:
            return resp.text

    # ----------------------------------------------------------------------
    def _handle_json_error(self, error, errorcode):
        errormessage = error.get("message")
        # handles case where message exists in the dictionary but is None
        if errormessage is None:
            errormessage = "Unknown Error"
        # _log.error(errormessage)
        if "details" in error and error["details"] is not None:
            if isinstance(error["details"], str):
                errormessage = f"{errormessage} \n {error['details']}"
                # _log.error(error['details'])
            else:
                for errordetail in error["details"]:
                    if isinstance(errordetail, str):
                        errormessage = errormessage + "\n" + errordetail
                        # _log.error(errordetail)

        errormessage = errormessage + "\n(Error Code: " + str(errorcode) + ")"
        raise Exception(errormessage)

    def post_multipart(
        self, path: str, params: dict = None, files: list = None, **kwargs
    ) -> Union[dict, requests.Response]:
        """
        sends a MultiPart Form POST request.

        ===========================   =====================================================
        **Parameter**                **Description**
        ---------------------------   -----------------------------------------------------
        path                          optional string.  URL or part of the url resource
                                      to call.
        ---------------------------   -----------------------------------------------------
        params                        optional dict.  Contains data to pass to the web
                                      resource.
        ---------------------------   -----------------------------------------------------
        files                         optional list of files

                                      Files can be provided two ways:

                                      The most basic way is:
                                      Way1: {key : r"c:\temp\myfile.foo}
                                      This is just the file path and the key.


                                      The preferred way:

                                      Way 2: {key : (file_name, open(c:\temp\myfile.foo, 'rb'), image\jpeg)}

                                      Way 2 requires providing the filename, IO object as 'rb', and the mimetype.
        ===========================   =====================================================


        ===========================   =====================================================
        **Optional Parameters**       **Description**
        ---------------------------   -----------------------------------------------------
        add_token                     optional boolean.  True means try to add the boolean,
                                      else do not add a ?token=<foo> to the call. If
                                      is_geoevent is True, then the token will be appended
                                      to the header.
        ---------------------------   -----------------------------------------------------
        is_geoevent                   optional boolean. True means the auth token will be placed in the header
        ---------------------------   -----------------------------------------------------
        try_json                      optional boolean.  If true, the call adds the ?f=json.
        ---------------------------   -----------------------------------------------------
        ssl                           optional boolean. If true all calls are forced to be
                                      https.
        ---------------------------   -----------------------------------------------------
        out_folder                    optional string.  This is the save folder for the data.
        ---------------------------   -----------------------------------------------------
        file_name                     optional string. The save name of the file. This will override the file name if provided in the response.
        ---------------------------   -----------------------------------------------------
        force_bytes                   optional boolean.  Deprecated.
        ---------------------------   -----------------------------------------------------
        add_headers                   optional dict.  If provided, additional headers will be given for a single call.
        ---------------------------   -----------------------------------------------------
        post_json                     optional bool. If True, the data is pushed in the request's json parameter.  This is an edge case for Workflow Manager. The default is `False`.
        ---------------------------   -----------------------------------------------------
        json_encode                   optional Bool. If False, the key/value parameters will not be JSON encoded.
        ---------------------------   -----------------------------------------------------
        timeout                       optional Integer. Timeout in seconds
        ---------------------------   -----------------------------------------------------
        return_raw_response           Optional Boolean. If True, returns the requests.Response object.
        ---------------------------   -----------------------------------------------------
        allow_redirects               Optional Boolean. Does not allow the call to follow the 3xx status codes. The default is True
        ---------------------------   -----------------------------------------------------
        drop_auth                     Optional Boolean. Drop this when auth handlers cause issues. The default is False
        ===========================   =====================================================

        :returns: data returned from the URL call.
        """
        import io

        drop_auth = kwargs.pop("drop_auth", False)
        allow_redirects = kwargs.pop("allow_redirects", True)
        timeout = kwargs.pop("timeout", self._timeout)
        return_raw_response = kwargs.pop("return_raw_response", False)
        json_encode = kwargs.pop("json_encode", True)
        if self._baseurl.endswith("/") == False:
            self._baseurl += "/"
        url = path

        post_json = kwargs.pop("post_json", False)

        # if self._auth == "IWA":
        #    self._session = None
        if "postdata" in kwargs:  # handles legacy issues
            params = kwargs.pop("postdata")
        if params is None:
            params = {}
        if self._session is None:
            self._create_session()
        try_json = kwargs.pop("try_json", True)

        if url.find("://") == -1:
            url = self._baseurl + url
        if kwargs.pop("ssl", False) or self._all_ssl:
            url = url.replace("http://", "https://")

        if try_json:
            params["f"] = "json"
        fields = {}
        if files:
            if isinstance(files, dict):
                for k, v in files.items():
                    if isinstance(v, (list, tuple)):
                        fields[k] = v
                    else:
                        fields[k] = (
                            os.path.basename(v),
                            open(v, "rb"),
                            mimetypes.guess_type(v)[0],
                        )
            elif isinstance(files, (list, tuple)):
                for key, filePath, fileName in files:
                    if (
                        isinstance(fileName, str)
                        and isinstance(filePath, (io.StringIO, io.BytesIO)) == False
                    ):
                        fields[key] = (
                            fileName,
                            open(filePath, "rb"),
                            mimetypes.guess_type(filePath)[0],
                        )
                    elif isinstance(fileName, str) and isinstance(
                        filePath, (io.StringIO, io.BytesIO)
                    ):
                        fields[key] = (
                            fileName,
                            filePath,
                            None,
                        )
                    else:
                        fields[key] = v
            files = fields

        out_path = (
            kwargs.pop("out_path", None)
            or kwargs.pop("out_folder", None)
            or tempfile.gettempdir()
        )
        file_name = kwargs.pop("file_name", None)
        try:
            if self._cert_file:
                cert = (self._cert_file, self._key_file)
            else:
                cert = None
            if json_encode:
                for k, v in params.items():
                    if isinstance(v, (dict, list, tuple, bool)):
                        params[k] = json.dumps(v, default=_date_handler)
                    elif isinstance(v, PropertyMap):
                        params[k] = json.dumps(dict(v), default=_date_handler)
                    elif isinstance(v, InsensitiveDict):
                        params[k] = v.json
            # When data and files are present, they need to be combined
            # https://stackoverflow.com/a/12385661

            params.update(fields)
            if self._session.auth and drop_auth:
                auth = self._session.auth
                self._session.auth
            else:
                auth = None
            if post_json:  # edge case workflow
                if timeout:
                    resp = self._session.post(
                        url=url,
                        json=params,
                        cert=cert,
                        files=files,
                        allow_redirects=allow_redirects,
                        verify=self._verify_cert,
                        timeout=timeout,
                    )
                else:
                    resp = self._session.post(
                        url=url,
                        json=params,
                        cert=cert,
                        allow_redirects=allow_redirects,
                        verify=self._verify_cert,
                        files=files,
                    )
            else:
                if files is None:
                    files = {}
                multipart_form_data = {
                    k: (None, v) for k, v in params.items() if not k in files
                }
                multipart_form_data.update(fields)

                if timeout:
                    resp = self._session.post(
                        url=url,
                        files=multipart_form_data,
                        cert=cert,
                        allow_redirects=allow_redirects,
                        timeout=timeout,
                        verify=self._verify_cert,
                    )
                else:
                    resp = self._session.post(
                        url=url,
                        cert=cert,
                        files=multipart_form_data,
                        allow_redirects=allow_redirects,
                        verify=self._verify_cert,
                    )
            if auth and drop_auth:
                self._session.auth = auth
        except requests.exceptions.SSLError as err:
            raise requests.exceptions.SSLError(
                "Please set verify_cert=False due to encountered SSL error: %s" % err
            )
        except requests.exceptions.InvalidURL as errIU:
            raise requests.exceptions.SSLError("Invalid URL provided: %s" % errIU)
        except requests.exceptions.ConnectionError as errCE:
            raise requests.exceptions.ConnectionError(
                "A connection error has occurred: %s" % errCE
            )

        except requests.exceptions.InvalidHeader as errIH:
            raise requests.exceptions.InvalidHeader(
                "A invalid header was provided: %s" % errIH
            )
        except requests.exceptions.HTTPError as errh:
            raise requests.exceptions.HTTPError("Http Error: %s" % errh)
        except requests.exceptions.MissingSchema as errMS:
            raise requests.exceptions.MissingSchema(
                "URL scheme must be provided: %s" % errMS
            )
        except requests.exceptions.RequestException as errRE:
            raise requests.exceptions.RequestException(
                "A general exception was raised: %s" % errRE
            )
        except Exception as e:
            raise Exception("A general error occurred: %s" % e)
        except:
            import traceback

            raise Exception("An unknown error occurred: %s" % traceback.format_exc())
        if return_raw_response:
            return resp
        return self._handle_response(
            resp=resp,
            out_path=out_path,
            file_name=file_name,
            try_json=try_json,
            force_bytes=kwargs.pop("force_bytes", False),
        )

    # ----------------------------------------------------------------------
    def post(
        self, path: str, params: dict = None, files: list = None, **kwargs
    ) -> Union[dict, requests.Response]:
        """
        sends a POST request.

        ===========================   =====================================================
        **Parameter**                **Description**
        ---------------------------   -----------------------------------------------------
        path                          optional string.  URL or part of the url resource
                                      to call.
        ---------------------------   -----------------------------------------------------
        params                        optional dict.  Contains data to pass to the web
                                      resource.
        ---------------------------   -----------------------------------------------------
        files                         optional list of files

                                      Files can be provided two ways:

                                      The most basic way is:
                                      Way1: {key : r"c:\temp\myfile.foo}
                                      This is just the file path and the key.


                                      The preferred way:

                                      Way 2: {key : (file_name, open(c:\temp\myfile.foo, 'rb'), image\jpeg)}

                                      Way 2 requires providing the filename, IO object as 'rb', and the mimetype.
        ===========================   =====================================================


        ===========================   =====================================================
        **Optional Parameters**       **Description**
        ---------------------------   -----------------------------------------------------
        add_token                     optional boolean.  True means try to add the boolean,
                                      else do not add a ?token=<foo> to the call. If
                                      is_geoevent is True, then the token will be appended
                                      to the header.
        ---------------------------   -----------------------------------------------------
        is_geoevent                   optional boolean. True means the auth token will be placed in the header
        ---------------------------   -----------------------------------------------------
        try_json                      optional boolean.  If true, the call adds the ?f=json.
        ---------------------------   -----------------------------------------------------
        ssl                           optional boolean. If true all calls are forced to be
                                      https.
        ---------------------------   -----------------------------------------------------
        out_folder                    optional string.  This is the save folder for the data.
        ---------------------------   -----------------------------------------------------
        file_name                     optional string. The save name of the file. This will override the file name if provided in the response.
        ---------------------------   -----------------------------------------------------
        force_bytes                   optional boolean.  Deprecated.
        ---------------------------   -----------------------------------------------------
        add_headers                   optional dict.  If provided, additional headers will be given for a single call.
        ---------------------------   -----------------------------------------------------
        post_json                     optional bool. If True, the data is pushed in the request's json parameter.  This is an edge case for Workflow Manager. The default is `False`.
        ---------------------------   -----------------------------------------------------
        json_encode                   optional Bool. If False, the key/value parameters will not be JSON encoded.
        ---------------------------   -----------------------------------------------------
        timeout                       optional Integer. The number of seconds to timeout a service without a response.  The default is 600 seconds.
        ---------------------------   -----------------------------------------------------
        return_raw_response           Optional boolean. Returns the requests.Response object.
        ---------------------------   -----------------------------------------------------
        allow_redirects               Optional Boolean. Does not allow the call to follow the 3xx status codes. The default is True
        ---------------------------   -----------------------------------------------------
        drop_auth                     Optional Boolean. Drop this when auth handlers cause issues. The default is False
        ===========================   =====================================================

        :returns: data returned from the URL call.

        """
        drop_auth = kwargs.pop("drop_auth", False)
        allow_redirects = kwargs.pop("allow_redirects", True)
        timeout = kwargs.pop("timeout", self._timeout)
        return_raw_response = kwargs.pop("return_raw_response", False)
        json_encode = kwargs.pop("json_encode", True)
        add_headers = kwargs.pop("add_headers", {})
        buffer_reader = None

        if self._baseurl.endswith("/") == False:
            self._baseurl += "/"
        url = path

        post_json = kwargs.pop("post_json", False)

        if "postdata" in kwargs:  # handles legacy issues
            params = kwargs.pop("postdata")
        if params is None:
            params = {}
        if self._session is None:
            self._create_session()
        try_json = kwargs.pop("try_json", True)

        if url.find("://") == -1:
            url = self._baseurl + url
        if kwargs.pop("ssl", False) or self._all_ssl:
            url = url.replace("http://", "https://")

        if try_json:
            params["f"] = "json"
        if files:
            fields = {}
            if isinstance(files, dict):
                for k, v in files.items():
                    if isinstance(v, (list, tuple)):
                        fields[k] = v
                    else:
                        buffer_reader = open(v, "rb")
                        fields[k] = (
                            os.path.basename(v),
                            buffer_reader,
                            mimetypes.guess_type(v)[0],
                        )
            elif isinstance(files, (list, tuple)):
                for key, filePath, fileName in files:
                    import io

                    if (
                        isinstance(fileName, str)
                        and isinstance(filePath, (io.StringIO, io.BytesIO)) == False
                    ):
                        buffer_reader = open(filePath, "rb")
                        fields[key] = (
                            fileName,
                            buffer_reader,
                            mimetypes.guess_type(filePath)[0],
                        )
                    elif isinstance(fileName, str) and isinstance(
                        filePath, (io.StringIO, io.BytesIO)
                    ):
                        fields[key] = (
                            fileName,
                            filePath,
                            None,
                        )
                    else:
                        fields[key] = v
            files = fields

        out_path = (
            kwargs.pop("out_path", None)
            or kwargs.pop("out_folder", None)
            or tempfile.gettempdir()
        )
        file_name = kwargs.pop("file_name", None)
        if add_headers:
            original_headers = copy.deepcopy(self._session.headers)
            self._session.headers.update(add_headers)
        try:
            if self._cert_file:
                cert = (self._cert_file, self._key_file)
            else:
                cert = None
            if json_encode:
                for k, v in params.items():
                    if isinstance(v, (dict, list, tuple, bool)):
                        params[k] = json.dumps(v, default=_date_handler)
                    elif isinstance(v, PropertyMap):
                        params[k] = json.dumps(dict(v), default=_date_handler)
                    elif isinstance(v, InsensitiveDict):
                        params[k] = v.json
            if self._session.auth and drop_auth:
                auth = self._session.auth
                self._session.auth
            else:
                auth = None
            if post_json:  # edge case workflow
                if timeout:
                    resp = self._session.post(
                        url=url,
                        json=params,
                        cert=cert,
                        files=files,
                        timeout=timeout,
                        verify=self._verify_cert,
                        allow_redirects=allow_redirects,
                    )
                else:
                    resp = self._session.post(
                        url=url,
                        json=params,
                        cert=cert,
                        files=files,
                        verify=self._verify_cert,
                        allow_redirects=allow_redirects,
                    )

            else:
                if timeout:
                    resp = self._session.post(
                        url=url,
                        data=params,
                        cert=cert,
                        files=files,
                        timeout=timeout,
                        verify=self._verify_cert,
                        allow_redirects=allow_redirects,
                    )
                else:
                    resp = self._session.post(
                        url=url,
                        data=params,
                        cert=cert,
                        files=files,
                        verify=self._verify_cert,
                        allow_redirects=allow_redirects,
                    )
            if auth:
                self._session.auth = auth
            if add_headers:
                self._session.headers.clear()
                self._session.headers.update(original_headers)
        except requests.exceptions.SSLError as err:
            raise requests.exceptions.SSLError(
                "Please set verify_cert=False due to encountered SSL error: %s" % err
            )
        except requests.exceptions.InvalidURL as errIU:
            raise requests.exceptions.SSLError("Invalid URL provided: %s" % errIU)
        except requests.exceptions.ConnectionError as errCE:
            raise requests.exceptions.ConnectionError(
                "A connection error has occurred: %s" % errCE
            )

        except requests.exceptions.InvalidHeader as errIH:
            raise requests.exceptions.InvalidHeader(
                "A invalid header was provided: %s" % errIH
            )
        except requests.exceptions.HTTPError as errh:
            raise requests.exceptions.HTTPError("Http Error: %s" % errh)
        except requests.exceptions.MissingSchema as errMS:
            raise requests.exceptions.MissingSchema(
                "URL scheme must be provided: %s" % errMS
            )
        except requests.exceptions.RequestException as errRE:
            raise requests.exceptions.RequestException(
                "A general exception was raised: %s" % errRE
            )
        except Exception as e:
            raise Exception("A general error occurred: %s" % e)
        except:
            import traceback

            raise Exception("An unknown error occurred: %s" % traceback.format_exc())
        if buffer_reader:
            buffer_reader.close()
        if return_raw_response:
            return resp
        return self._handle_response(
            resp=resp,
            out_path=out_path,
            file_name=file_name,
            try_json=try_json,
            force_bytes=kwargs.pop("force_bytes", False),
        )

    # ----------------------------------------------------------------------
    def put_raw(self, url: str, data: dict, **kwargs) -> requests.Response:
        """
        performs a raw PUT operation

        url: str
        data: bytes or open() object
        kwargs - optional requests.put parameters.  headers is not supported, use additional_headers
        """
        verify = kwargs.pop("verify", True)
        original_headers = copy.deepcopy(self._session.headers)
        self._session.headers.update(kwargs.pop("additional_headers", {}))
        token_header = "X-Esri-Authorization"
        if self.token and not "X-Esri-Authorization" in original_headers:
            token = self.token
            self._session.headers.update({token_header: "Bearer %s" % token})

        resp = self._session.put(
            url=url,
            data=data,
            verify=verify,
            headers=self._session.headers,
            **kwargs,
        )
        self._session.headers.clear()
        self._session.headers.update(original_headers)

        return resp

    # ----------------------------------------------------------------------
    def put(
        self, url: str, params: dict = None, files: list = None, **kwargs
    ) -> Union[dict, requests.Response]:
        """
        sends a PUT request

        ===========================   =====================================================
        **Optional Parameters**       **Description**
        ---------------------------   -----------------------------------------------------
        url                           Optional String. The web endpoint.
        ---------------------------   -----------------------------------------------------
        params                        Optional dict. The Key/value pairs to send along with the request.
        ---------------------------   -----------------------------------------------------
        files                         Optional list. Allows users to provide a file or files
                                      to the operation.  The format should be:
                                      files = [[key, filePath, fileName], ...,[keyN, filePathN, fileNameN]]
        ===========================   =====================================================

        ===========================   =====================================================
        **Optional Parameters**       **Description**
        ---------------------------   -----------------------------------------------------
        add_token                     optional boolean.  True means try to add the boolean,
                                      else do not add a ?token=<foo> to the call. (deprecated)
        ---------------------------   -----------------------------------------------------
        token_as_header               Optional boolean.  If True, the token will go into
                                      the header as `Authorization` header.  This can be
                                      overwritten using the `token_header` parameter. (deprecated)
        ---------------------------   -----------------------------------------------------
        token_header                  Optional String. If provided and token_as_header is
                                      True, authentication token will be placed in this
                                      instead on URL string. (deprecated)
        ---------------------------   -----------------------------------------------------
        try_json                      optional boolean.  If true, the call adds the ?f=json.
        ---------------------------   -----------------------------------------------------
        ssl                           optional boolean. If true all calls are forced to be
                                      https.
        ---------------------------   -----------------------------------------------------
        post_json                     optional bool. If True, the data is pushed in the request's json parameter.  This is an edge case for Workflow Manager. The default is `False`.
        ---------------------------   -----------------------------------------------------
        json_encode                   optional Bool. If False, the key/value parameters will not be JSON encoded.
        ---------------------------   -----------------------------------------------------
        return_raw_response           Optional boolean. When true, it returns the requests.Response object
        ---------------------------   -----------------------------------------------------
        allow_redirects               Optional Boolean. Does not allow the call to follow the 3xx status codes. The default is True
        ---------------------------   -----------------------------------------------------
        drop_auth                     Optional Boolean. Drop this when auth handlers cause issues. The default is False
        ===========================   =====================================================

        :returns: dict or string depending on the response

        """
        drop_auth = kwargs.pop("drop_auth", False)
        allow_redirects = kwargs.pop("allow_redirects", True)
        return_raw_response = kwargs.pop("return_raw_response", False)
        post_json = kwargs.pop("post_json", False)
        json_encode = kwargs.pop("json_encode", True)
        out_path = kwargs.pop("out_path", None)
        file_name = kwargs.pop("file_name", None)
        if params is None:
            params = {}
        if self._session is None:
            self._create_session()
        try_json = kwargs.pop("try_json", True)
        if url.find("://") == -1:
            url = self._baseurl + url
        if kwargs.pop("ssl", False):
            url = url.replace("http://", "https://")

        if try_json:
            params["f"] = "json"

        if files:
            fields = {}
            if isinstance(files, dict):
                for k, v in files.items():
                    if isinstance(v, (list, tuple)):
                        fields[k] = v
                    else:
                        fields[k] = (
                            os.path.basename(v),
                            open(v, "rb"),
                            mimetypes.guess_type(v)[0],
                        )
            elif isinstance(files, (list, tuple)):
                for key, filePath, fileName in files:
                    if isinstance(fileName, str):
                        fields[key] = (
                            fileName,
                            open(filePath, "rb"),
                            mimetypes.guess_type(filePath)[0],
                        )
                    else:
                        fields[key] = v

            params.update(fields)
            del files, fields
        if self._cert_file:
            cert = (self._cert_file, self._key_file)
        else:
            cert = None
        if json_encode:
            for k, v in params.items():
                if isinstance(v, (dict, list, tuple, bool)):
                    params[k] = json.dumps(v, default=_date_handler)
                elif isinstance(v, PropertyMap):
                    params[k] = json.dumps(dict(v), default=_date_handler)
                elif isinstance(v, InsensitiveDict):
                    params[k] = v.json
        if self._session.auth and drop_auth:
            auth = self._session.auth
            self._session.auth
        else:
            auth = None
        if post_json:  # edge case workflow
            resp = self._session.put(
                url=url,
                json=params,
                cert=cert,
                files=files,
                allow_redirects=allow_redirects,
            )
        else:
            resp = self._session.put(
                url=url,
                data=params,
                cert=cert,
                files=files,
                allow_redirects=allow_redirects,
            )
        if auth and drop_auth:
            self._session.auth = auth
        if return_raw_response:
            return resp
        return self._handle_response(
            resp=resp,
            out_path=out_path,
            file_name=file_name,
            try_json=try_json,
            force_bytes=kwargs.pop("force_bytes", False),
        )

    # ----------------------------------------------------------------------
    def delete(self, url, params=None, **kwargs):
        """
        sends a PUT request

        ===========================   =====================================================
        **Parameter**                **Description**
        ---------------------------   -----------------------------------------------------
        url                           Optional String. The web endpoint.
        ---------------------------   -----------------------------------------------------
        params                        Optional dict. The Key/value pairs to send along with the request.
        ===========================   =====================================================

        ===========================   =====================================================
        **Optional Parameters**       **Description**
        ---------------------------   -----------------------------------------------------
        add_token                     optional boolean.  True means try to add the boolean,
                                      else do not add a ?token=<foo> to the call.
        ---------------------------   -----------------------------------------------------
        token_as_header               Optional boolean.  If True, the token will go into
                                      the header as `Authorization` header.  This can be
                                      overwritten using the `token_header` parameter
        ---------------------------   -----------------------------------------------------
        token_header                  Optional String. If provided and token_as_header is
                                      True, authentication token will be placed in this
                                      instead on URL string.
        ---------------------------   -----------------------------------------------------
        try_json                      optional boolean.  If true, the call adds the ?f=json.
        ---------------------------   -----------------------------------------------------
        ssl                           optional boolean. If true all calls are forced to be
                                      https.
        ---------------------------   -----------------------------------------------------
        return_raw_response           Optional Boolean.  Returns the raw requests.Response object
        ---------------------------   -----------------------------------------------------
        allow_redirects               Optional Boolean. Does not allow the call to follow the 3xx status codes. The default is True
        ---------------------------   -----------------------------------------------------
        drop_auth                     Optional Boolean. Drop this when auth handlers cause issues. The default is False
        ===========================   =====================================================

        :returns: dict or string depending on the response
        """
        allow_redirects = kwargs.pop("allow_redirects", True)
        drop_auth = kwargs.pop("drop_auth", False)
        out_path = kwargs.pop("out_path", None)
        return_raw_response = kwargs.pop("return_raw_response", False)
        file_name = kwargs.pop("file_name", None)
        if params is None:
            params = {}
        if self._session is None:
            self._create_session()
        try_json = kwargs.pop("try_json", True)
        if url.find("://") == -1:
            url = self._baseurl + url
        if kwargs.pop("ssl", False):
            url = url.replace("http://", "https://")

        if try_json:
            params["f"] = "json"
        if drop_auth and self._session.auth:
            auth = self._session.auth
            self._session.auth = None
            resp = self._session.delete(
                url=url, data=params, allow_redirects=allow_redirects
            )
            self._session.auth = auth
        else:
            resp = self._session.delete(
                url=url, data=params, allow_redirects=allow_redirects
            )
        if return_raw_response:
            return resp
        return self._handle_response(
            resp=resp,
            out_path=out_path,
            file_name=file_name,
            try_json=try_json,
            force_bytes=kwargs.pop("force_bytes", False),
        )

    # ----------------------------------------------------------------------
    def streaming_method(
        self,
        url: str,
        callback: object,
        data: dict = None,
        json_data: dict = None,
        verb: str = "GET",
        **kwargs,
    ) -> Union[dict, requests.Response]:
        """
        Performs streaming web requests.

        =======================     ===========================================================
        **Parameter**              **Description**
        -----------------------     -----------------------------------------------------------
        url                         Required String. The web resource location.
        -----------------------     -----------------------------------------------------------
        callback                    Required Method.  The callback function to handle the response from the streaming request.

                                    **Example**
                                    ```
                                    def hook(r, *args, **kwargs):
                                        print('called a hook')
                                        return r
                                    ```
                                    See: https://requests.kennethreitz.org/en/master/user/advanced/#event-hooks


        -----------------------     -----------------------------------------------------------
        data                        Optional Dict. The parameters to pass to the method.
        -----------------------     -----------------------------------------------------------
        json_data                   Optional Dict. The parameters to pass to the method. This applies to POST only
        -----------------------     -----------------------------------------------------------
        verb                        Optional String.  The default is GET.  The allowed values are POST, GET, or PUT.
        -----------------------     -----------------------------------------------------------
        kwargs                      Optional Dict.  See https://requests.readthedocs.io/en/master/user/advanced/#request-and-response-objects
        =======================     ===========================================================

        """
        verbs = ["post", "put", "get"]
        if verb.lower() in verbs:
            hooks = {"response": callback}
            fn = getattr(self._session, verb.lower())
            if verb.lower() == "post":
                return fn(
                    url=url,
                    data=data,
                    json_data=json,
                    hooks=hooks,
                    stream=True,
                    **kwargs,
                )
            else:
                return fn(url=url, data=data, hooks=hooks, stream=True, **kwargs)
        else:
            allowed_verb = ",".join(verbs)
            raise ValueError(f"Invalid web method only {allowed_verb} as allowed")
        self._session.post(url=url, data=data, json_data=json, stream=True)

    # ----------------------------------------------------------------------
    def login(
        self,
        username: str,
        password: str,
        expiration: Union[int, float] = None,
    ):
        """allows a user to login to a site with different credentials"""
        if expiration is None:
            expiration = 1440
        try:
            if self._username != username and self._password != password:
                c = Connection(
                    baseurl=self._baseurl,
                    username=username,
                    password=password,
                )
                self = c
        except:
            raise Exception("Could not create a new login.")

    # ----------------------------------------------------------------------
    def relogin(self, expiration: Union[float, int] = None) -> str:
        """Re-authenticates with the portal using the same username/password."""
        if expiration is None:
            expiration = self._expiration
        self.logout()
        return self.token

    # ----------------------------------------------------------------------
    def logout(self):
        """Logs out of the portal."""
        self._token = None
        self._create_time = None

    # ----------------------------------------------------------------------
    @property
    def is_logged_in(self) -> bool:
        """Returns true if logged into the portal."""
        return (self._auth in ["ANON", "UNKNOWN"]) == False

    # ----------------------------------------------------------------------
    @property
    def product(self) -> str:
        """Returns true if logged into the portal."""
        return self._product

    # ----------------------------------------------------------------------
    @property
    def token(self) -> str:
        """Gets a Token"""
        if isinstance(self._session.auth, EsriBuiltInAuth):
            return self._session.auth.token
        elif isinstance(self._session.auth, EsriGenTokenAuth):
            return self._session.auth.token()
        elif isinstance(self._session.auth, ArcGISProAuth):
            return self._session.auth.token
        elif isinstance(self._session.auth, EsriOAuth2Auth):
            return self._session.auth.token
        elif isinstance(self._session.auth, EsriUserTokenAuth):
            return self._session.auth.token
        elif isinstance(self._session.auth, EsriNotebookAuth):
            return self._session.auth.token
        elif isinstance(self._session.auth, EsriAPIKeyAuth):
            return self._session.auth.token
        return None

    # ----------------------------------------------------------------------
    @token.setter
    def token(self, value):
        """gets/sets the token"""
        if self._token != value:
            self._token = value

    # ----------------------------------------------------------------------
    def _check_product(self) -> str:
        """
        determines if the product is portal, arcgis online or arcgis server
        """
        from urllib.error import HTTPError

        baseurl = self._baseurl
        if self._product == "SERVER":
            if self._token_url is None:
                parsed = urlparse(self._baseurl)
                path = parsed.path[1:].split("/")[0]
                self._token_url = "https://%s/%s/admin/generateToken" % (
                    parsed.netloc,
                    path,
                )
            self._product = "SERVER"
            return "SERVER"
        if baseurl is None:
            return "UNKNOWN"
        if baseurl.lower().find("arcgis.com") > -1:
            parsed = urlparse(self._baseurl)
            self._product = "AGOL"
            self._token_url = "https://%s/sharing/rest/generateToken" % parsed.netloc
            return "AGOL"
        elif baseurl.lower().find("/sharing/rest") > -1:
            if baseurl.endswith("/"):
                try:
                    res = self.get(
                        baseurl + "info",
                        params={"f": "json"},
                        add_token=False,
                    )
                except:
                    res = self.get(baseurl + "info", params={"f": "json"})
            else:
                try:
                    res = self.get(
                        baseurl + "/info",
                        params={"f": "json"},
                        add_token=False,
                    )
                except:
                    res = self.get(baseurl + "/info", params={"f": "json"})
            if (
                self._token_url is None
                and res is not None
                and isinstance(res, dict)
                and "authInfo" in res
                and "tokenServicesUrl" in res["authInfo"]
                and res["authInfo"]["isTokenBasedSecurity"]
            ):
                parsed_from_system = urlparse(res["authInfo"]["tokenServicesUrl"])
                parsed = urlparse(baseurl)
                if (
                    parsed.netloc.lower() != parsed_from_system.netloc.lower()
                    and parsed.netloc.find(":7443") > -1
                ):  # WA not being used for token url
                    self._token_url = os.path.join(
                        parsed_from_system.scheme + "://",
                        parsed.netloc
                        + "/arcgis/"
                        + "/".join(parsed_from_system.path[1:].split("/")[1:]),
                    )
                    url_test = self._session.post(
                        self._token_url, {"f": "json"}, allow_redirects=False
                    )
                    if url_test.status_code == 301:
                        self._token_url = url_test.headers["location"]
                        if self._baseurl != os.path.dirname(
                            url_test.headers["location"]
                        ):
                            self._baseurl = os.path.dirname(
                                url_test.headers["location"]
                            )
                else:
                    self._token_url = res["authInfo"]["tokenServicesUrl"]
            elif (
                self._token_url is None
                and res is not None
                and isinstance(res, dict)
                and "authInfo" in res
                and "tokenServicesUrl" in res["authInfo"]
                and res["authInfo"]["isTokenBasedSecurity"] == False
            ):
                self._token_url = None
                self._auth = "OTHER"
            self._product = "PORTAL"
            return "PORTAL"
        else:
            # Brute Force Method
            parsed = urlparse(baseurl)
            root = (
                rf"{parsed.scheme}://{parsed.netloc}/{parsed.path[1:].split(r'/')[0]}"
            )
            parts = [
                "/info",
                "/rest/services",
                "/rest/info",
                "/sharing/rest/info",
            ]
            params = {"f": "json"}
            for pt in parts:
                try:
                    res = self.get(
                        root + pt,
                        params=params,
                        add_token=False,
                        allow_redirects=False,
                    )
                    if "folders" in res:
                        self._product = "SERVER"
                        return self._check_product()
                    elif (
                        self._token_url is None
                        and res is not None
                        and isinstance(res, dict)
                        and "authInfo" in res
                        and "tokenServicesUrl" in res["authInfo"]
                        and res["authInfo"]["isTokenBasedSecurity"]
                    ):
                        self._token_url = res["authInfo"]["tokenServicesUrl"]
                    elif (
                        self._token_url is None
                        and res is not None
                        and isinstance(res, dict)
                        and "authInfo" in res
                        and "tokenServicesUrl" in res["authInfo"]
                        and res["authInfo"]["isTokenBasedSecurity"] == False
                    ):
                        self._token_url = None
                        self._auth = "OTHER"
                except HTTPError as e:
                    _log.warning(str(e))
                    res = ""
                except json.decoder.JSONDecodeError:
                    res = ""
                except Exception as e:
                    res = ""
                if (
                    isinstance(res, dict)
                    and "currentVersion" in res
                    and self._token_url
                ):
                    t_parsed = urlparse(self._token_url[1:]).path
                    b_parsed = urlparse(self._baseurl[1:]).path
                    if t_parsed.startswith("/"):
                        t_parsed = t_parsed[1:].split("/")[0]
                    else:
                        t_parsed = t_parsed.split("/")[0]
                    if b_parsed.startswith("/"):
                        b_parsed = b_parsed[1:].split("/")[0]
                    else:
                        b_parsed = b_parsed.split("/")[0]
                    if t_parsed.lower() != b_parsed.lower():
                        self._token_url = None
                        if self._portal_connection:
                            self._product = "FEDERATED_SERVER"
                            return "FEDERATED_SERVER"
                        else:
                            from arcgis.gis import GIS

                            self._portal_connection = GIS(
                                url=res["authInfo"]["tokenServicesUrl"].split(
                                    "/sharing/"
                                )[0],
                                username=self._username,
                                password=self._password,
                                verify_cert=self._verify_cert,
                            )._con
                            self._product = "FEDERATED_SERVER"
                            return "FEDERATED_SERVER"
                    self._product = "SERVER"
                    return "SERVER"
                elif (
                    isinstance(res, dict)
                    and "currentVersion" in res
                    and self._token_url is None
                ):
                    self._product = "SERVER"
                    return "SERVER"
                del pt
                del res
        self._product = "PORTAL"
        return "PORTAL"

    def _create_token(self, url: str) -> str:
        """
        When a service needs a token that is not in the token, this method obtains it.
        Only the following authentications require a URL: EsriPKIAuth, EsriBasicAuth, EsriKerberosAuth, EsriWindowsAuth, or EsriGenTokenAuth
        :returns: str (token is possible)
        """
        from arcgis.auth import (
            EsriNotebookAuth,
            EsriAPIKeyAuth,
            EsriPKIAuth,
            EsriBasicAuth,
            EsriKerberosAuth,
            EsriWindowsAuth,
            BaseEsriAuth,
        )
        from arcgis.auth.tools import parse_url

        if isinstance(
            self._session.auth,
            (
                EsriUserTokenAuth,
                EsriOAuth2Auth,
                EsriNotebookAuth,
                EsriAPIKeyAuth,
                ArcGISProAuth,
                EsriBuiltInAuth,
            ),
        ):
            return self._session.auth.token
        elif isinstance(self._session.auth, EsriGenTokenAuth):
            return self._session.auth.token(url)
        elif isinstance(
            self._session.auth,
            (EsriPKIAuth, EsriBasicAuth, EsriKerberosAuth, EsriWindowsAuth),
        ):
            if self._server_log is None:
                self._server_log = {}
            parsed = parse_url(url)
            expiration = 16000
            if parsed.port:
                if parsed.port in parsed.netloc:
                    server_url = f'{parsed.scheme}://{parsed.netloc}/{parsed.path[1:].split("/")[0]}'
                else:
                    server_url = f'{parsed.scheme}://{parsed.netloc}:{parsed.port}/{parsed.path[1:].split("/")[0]}'
            else:
                server_url = (
                    f'{parsed.scheme}://{parsed.netloc}/{parsed.path[1:].split("/")[0]}'
                )
            postdata = {
                "request": "getToken",
                "serverURL": server_url,
                "referer": self._referer or "http",
                "f": "json",
            }

            if expiration:
                postdata["expiration"] = expiration
            if parsed.netloc in self._server_log:
                token_url = self._server_log[parsed.netloc]
            else:
                info = self._session.get(
                    server_url + "/rest/info?f=json",
                    auth=self._session.auth,
                    verify=self._verify_cert,
                ).json()
                token_url = info["authInfo"]["tokenServicesUrl"]
                self._server_log[parsed.netloc] = token_url

            token = self._session.post(
                token_url,
                data=postdata,
                auth=self._session.auth,
                verify=self._verify_cert,
            )
            return token.json().get("token", None)
        elif isinstance(self._session.auth, BaseEsriAuth):
            return self._session.auth.token
        return None



from arcgis.auth.tools import LazyLoader

try:
    arcpy = LazyLoader("arcpy", strict=True)

    hasArcPy = True
except:
    hasArcPy = False
tempfile = LazyLoader("tempfile")


# ----------------------------------------------------------------------
def scratchFolder():
    if hasArcPy:
        return arcpy.env.scratchFolder
    else:
        return tempfile.gettempdir()


# ----------------------------------------------------------------------
def scratchGDB():
    if hasArcPy:
        return arcpy.env.scratchGDB
    return tempfile.gettempdir()


# ----------------------------------------------------------------------
def json_to_featureclass(json_file, out_fc):
    if hasArcPy:
        return arcpy.JSONToFeatures_conversion(
            in_json_file=json_file, out_features=out_fc
        )[0]
    return None


# ----------------------------------------------------------------------



from __future__ import annotations
from arcgis.auth.tools._lazy import LazyLoader

from typing import Any
from dataclasses import dataclass, field
from enum import Enum

arcgis = LazyLoader("arcgis")
__all__ = ["TraversabilityScopeEnum", "FilterScopeEnum"]


###########################################################################
def _parse_enum(value: Enum | Any | None) -> Any | None:
    """returns the Enum's value or the current value"""
    if isinstance(value, Enum):
        return value.value
    else:
        return value


###########################################################################
class TraversabilityScopeEnum(Enum):
    JUNCTIONS = "junctions"
    EDGES = "edges"
    JUNCTIONS_AND_EDGES = "junctionsAndEdges"


###########################################################################
class FilterScopeEnum(Enum):
    JUNCTIONS = "junctions"
    EDGES = "edges"
    JUNCTIONS_AND_EDGES = "junctionsAndEdges"


###########################################################################
@dataclass
class TraceConfiguration:
    """
    ========================================        ==========================================================
    **Parameter**                                    **Description**
    ----------------------------------------        ----------------------------------------------------------
    domain_network_name                             Required string. Specifies the name of the domain network
                                                    where the trace is starting. This is required for
                                                    subnetwork-based traces.
    ----------------------------------------        ----------------------------------------------------------
    tier_name                                       Required string. Specifies the name of the tier where the
                                                    trace is starting. This is required for subnetwork-based
                                                    traces.
    ----------------------------------------        ----------------------------------------------------------
    shortest_path_network_attribute_name            Required string for a shortest path trace; otherwise,
                                                    it's optional. It specifies the network attribute name
                                                    used for determining cost when calculating the shortest path.
    ----------------------------------------        ----------------------------------------------------------
    propagators                                     Required list of dictionaries. This is an array of objects.
                                                    The default is null.

                                                    Syntax:

                                                        [
                                                            {
                                                            "networkAttributeName" : "<string>",
                                                            "substitutionAttributeName": "<string>",
                                                            "propagatorfunctionType" : "bitwiseAnd" | "min", | "max"
                                                            "operator : "equal" | "notEqual"
                                                                        | "greaterThan"
                                                                        | "greaterThanEqual"
                                                                        | "lessThan"
                                                                        | "lessThanEqual"
                                                                        | "includesTheValues”
                                                                        | "doesNotIncludeTheValues"
                                                                        | "includesAny"
                                                                        | "doesNotIncludeAny",
                                                            "value" : string (numeric),
                                                            "propagatedAttributeName": "<string>"
                                                            }
                                                        ]
    ----------------------------------------        ----------------------------------------------------------
    target_tier_name                                Optional string. Specifies the name of the tier where an
                                                    upstream or downstream trace ends.
    ----------------------------------------        ----------------------------------------------------------
    subnetwork_name                                 Optional string. Specifies the name of the subnetwork that
                                                    be traced. The starting points of the trace are the controllers
                                                    of this subnetwork.
    ----------------------------------------        ----------------------------------------------------------
    filter_bitset_network_attribute_name            Optional string used during a loops trace to only return
                                                    loops with the same bit set all around the loop. This is
                                                    used during upstream and downstream traces to ensure that
                                                    trace results include any bit that is set in the starting
                                                    points for the network attribute.
    ----------------------------------------        ----------------------------------------------------------
    traversability_scope                            Optional TraversibilityScopeEnum or string specifying the
                                                    network element types to which the condition, category, or
                                                    function barriers apply. The default is junctionsAndEdges.
    ----------------------------------------        ----------------------------------------------------------
    filter_scope                                    Optional FilterScopeEnum or string specifying the network
                                                    element types to which the filter barriers or filter
                                                    function barriers apply. The default is junctionsAndEdges.
    ----------------------------------------        ----------------------------------------------------------
    condition_barriers                              Optional list of dictionaries.Each dictionary represents
                                                    network attribute or category conditions that serve as
                                                    barriers (the default is null).
                                                    If is_specific_value is true, the network attribute is
                                                    compared to a specific value; otherwise, the network
                                                    attribute is compared to another network attribute.

                                                    Syntax:

                                                        [
                                                            {
                                                                "name" : <string>,
                                                                "operator" : "equal" | "notEqual"
                                                                        | "greaterThan"
                                                                        | "greaterThanEqual"
                                                                        | "lessThan"
                                                                        | "lessThanEqual"
                                                                        | "includesTheValues"
                                                                        | "doesNotIncludeTheValues"
                                                                        | "includesAny"
                                                                        | "doesNotIncludeAny",
                                                                "value" : <string>,
                                                                "combineUsingOr" : <true | false>,
                                                                "isSpecificValue" : <true | false>
                                                            }
                                                        ]
    ----------------------------------------        ----------------------------------------------------------
    function_barriers                               Optional list of dictionaries. Each dictionary represents
                                                    a function barrier.

                                                    Syntax:

                                                        [
                                                            {
                                                                "functionType" : "add" | "subtract" |
                                                                            "average" | "count" |
                                                                            "min" | "max",
                                                                "networkAttributeName" : <string>,
                                                                "operator" : "equal" | "notEqual"
                                                                        | "greaterThan"
                                                                        | "greaterThanEqual"
                                                                        | "lessThan"
                                                                        | "lessThanEqual"
                                                                        | "includesTheValues"
                                                                        | "doesNotIncludeTheValues"
                                                                        | "includesAny"
                                                                        | "doesNotIncludeAny",
                                                                "value" : <string>,
                                                                "useLocalValues":true | false
                                                            }
                                                        ]
    ----------------------------------------        ----------------------------------------------------------
    filter_barriers                                 Optional list of dictionaries. Each dictionary represents
                                                    network attribute or category conditions that serve as
                                                    barriers (the default is null). If is_specific_value is
                                                    true, the network attribute is compared to a specific value;
                                                    otherwise, the network attribute is compared to
                                                    another network attribute.

                                                    Syntax:

                                                        [
                                                            {
                                                                "name" : <string>,
                                                                "operator" : "equal" | "notEqual"
                                                                        | "greaterThan"
                                                                        | "greaterThanEqual"
                                                                        | "lessThan"
                                                                        | "lessThanEqual"
                                                                        | "includesTheValues"
                                                                        | "doesNotIncludeTheValues"
                                                                        | "includesAny"
                                                                        | "doesNotIncludeAny",
                                                                "value" : string (numeric),
                                                                "combineUsingOr" : <true | false>,
                                                                "isSpecificValue" : <true | false>
                                                            }
                                                        ]
    ----------------------------------------        ----------------------------------------------------------
    filter_function_barriers                        Optional list of dictionaries. Each dictionary represents
                                                    a filter function barrier.

                                                    Syntax:

                                                        [
                                                            {
                                                                "functionType" : "add" | "subtract" |
                                                                            "average" | "count" |
                                                                            "min" | "max",
                                                                "networkAttributeName" : <string>,
                                                                "operator" : "equal" | "notEqual"
                                                                        | "greaterThan"
                                                                        | "greaterThanEqual"
                                                                        | "lessThan"
                                                                        | "lessThanEqual"
                                                                        | "includesTheValues"
                                                                        | "doesNotIncludeTheValues"
                                                                        | "includesAny"
                                                                        | "doesNotIncludeAny",
                                                                "value" : string (numeric)
                                                            }
                                                        ]
    ----------------------------------------        ----------------------------------------------------------
    functions                                       Optional list of dictionaries. Each dictionary represents
                                                    a function. Each function can have an optional list of
                                                    network attribute conditions.

                                                    Syntax:

                                                        [
                                                            {
                                                                "functionType" : "add" | "subtract"
                                                                            | "average" | "count"
                                                                            | "min" | "max",
                                                                "networkAttributeName" : <string>,
                                                                "conditions":
                                                                [
                                                                {
                                                                    "name" : <string>,
                                                                    "type" : "networkAttribute" | "category",
                                                                    "operator" : "equal" | "notEqual"
                                                                        | "greaterThan"
                                                                        | "greaterThanEqual |
                                                                        | "lessThan"
                                                                        | "lessThanEqual"
                                                                        | "includesTheValues"
                                                                        | "doesNotIncludeTheValues"
                                                                        | "includesAny"
                                                                        | "doesNotIncludeAny",
                                                                    "value" : <string>,
                                                                    "combineUsingOr" : <true | false>,
                                                                    "isSpecificValue" : <true | false>
                                                                }
                                                            ]
                                                            }

                                                        ]
    ----------------------------------------        ----------------------------------------------------------
    nearest_neighbor                                Optional dictionary that specifies the parameters needed
                                                    for calculating nearest neighbors.

                                                    Syntax:

                                                        {
                                                            "count" : int
                                                            "costNetworkAttributeName" : string,
                                                            "nearestCategories" : array of string,
                                                            "nearestAssets"
                                                            [
                                                                {
                                                                "networkSourceId" : long,
                                                                "assetGroupCode" : long,
                                                                "assetTypeCode" : long
                                                                }
                                                            ]
                                                        }
    ----------------------------------------        ----------------------------------------------------------
    output_filters                                  Optional list of dictionaries specifying the output filter.

                                                    Syntax:

                                                        [
                                                            {
                                                                "networkSourceId" : long,
                                                                "assetGroupCode" : long,
                                                                "assetTypeCode" : long
                                                            }
                                                        ]
    ----------------------------------------        ----------------------------------------------------------
    output_conditions                               Optional list of dictionaries specifying the type of features
                                                    returned based on a network attribute or category.

                                                    Syntax:

                                                        [
                                                            {
                                                                "name": "<string>",
                                                                "type": "networkAttribute" | "category",
                                                                "operator": "equal" | "notEqual"
                                                                            | "greaterThan"
                                                                            | "greaterThanEqual"
                                                                            | "lessThan"
                                                                            | "lessThanEqual"
                                                                            | "includesTheValues"
                                                                            | "doesNotIncludeTheValues"
                                                                            | "includesAny"
                                                                            | "doesNotIncludeAny",
                                                                    "value": <string>,
                                                                "combineUsingOr": <true | false>,
                                                                "isSpecificValue": <true | false>
                                                            }
                                                        ]
    ----------------------------------------        ----------------------------------------------------------
    include_containers                              Optional property specifying whether to include containers
                                                    in the trace result. The default is false.
    ----------------------------------------        ----------------------------------------------------------
    include_content                                 Optional property specifying whether to include content
                                                    in the trace result. The default is false.
    ----------------------------------------        ----------------------------------------------------------
    include_structures                              Optional property specifying whether to include structures
                                                    in the trace result. The default is false.
    ----------------------------------------        ----------------------------------------------------------
    include_barriers                                Optional property specifying whether to include barriers
                                                    in the trace result. The default is true.
    ----------------------------------------        ----------------------------------------------------------
    validate_consistency                            Optional property specifying whether to validate the
                                                    consistency of the trace results. The default is true.
    ----------------------------------------        ----------------------------------------------------------
    validate_locatability                           Optional property specifying whether to validate whether
                                                    traversed junction or edge objects have the necessary
                                                    containment, attachment, or connectivity association in
                                                    their association hierarchy. The default is false.
    ----------------------------------------        ----------------------------------------------------------
    include_isolated                                Optional property specifying whether to include isolated
                                                    features for an isolation trace. The default is false.
    ----------------------------------------        ----------------------------------------------------------
    ignore_barriers_at_starting_points              Optional property specifying whether dynamic barriers in
                                                    the trace configuration are ignored for starting points.
                                                    This can be useful when performing an upstream protective
                                                    device trace using the discovered protective devices
                                                    (barriers) as starting points to find subsequent upstream
                                                    protective devices. The default is false.
    ----------------------------------------        ----------------------------------------------------------
    include_up_to_first_spatial_container           Optional property specifying whether to limit the containers
                                                    returned in the trace result. This property depends on the
                                                    include_containers property and no-ops if include_containers
                                                    is false. If includeContainers is true and this property
                                                    is true, containment associations up to and including the
                                                    first spatial container are returned; otherwise, all c
                                                    ontainment associations are returned. The default is false.
    ----------------------------------------        ----------------------------------------------------------
    allow_indeterminate_flow                        Optional property specifying whether network features
                                                    with indeterminate flow stop traversability or are included
                                                    in the trace results. This property is only honored when
                                                    running an upstream, downstream, or isolation trace.
    ========================================        ==========================================================
    """

    domain_network_name: str
    tier_name: str
    shortest_path_network_attribute_name: str | None = None
    propagators: list[dict] | None = None
    target_tier_name: str | None = None
    subnetwork_name: str | None = None
    filter_bitset_network_attribute_name: str | None = None
    traversability_scope: TraversabilityScopeEnum | str | None = (
        TraversabilityScopeEnum.JUNCTIONS_AND_EDGES
    )
    filter_scope: FilterScopeEnum | str | None = FilterScopeEnum.JUNCTIONS_AND_EDGES
    condition_barriers: list[dict] | None = None
    function_barriers: list[dict] | None = None
    filter_barriers: list[dict] | None = None
    filter_function_barriers: list[dict] | None = None
    functions: list[dict] | None = None
    nearest_neighbor: dict | None = None
    output_filters: list[dict] | None = None
    output_conditions: list[dict] | None = None
    include_containers: bool = False
    include_content: bool = False
    include_structures: bool = False
    include_barriers: bool = True
    validate_consistency: bool = True
    validate_locatability: bool = False
    include_isolated: bool = False
    ignore_barriers_at_starting_points: bool = False
    include_up_to_first_spatial_container: bool = False
    allow_indeterminate_flow: bool | None = None
    _dict_data: dict | None = field(init=False)

    def __str__(self):
        return f"<TraceConfiguration: Domain Network Name= {self.domain_network_name}, Tier Name= {self.tier_name}>"

    def __repr__(self):
        return self.__str__()

    def __post_init__(self):
        self._dict_data = {
            "domainNetworkName": self.domain_network_name,
            "tierName": self.tier_name,
            "shortestPathNetworkAttributeName": self.shortest_path_network_attribute_name
            or "",
            "propagators": self.propagators or [],
            "targetTierName": self.target_tier_name or "",
            "subnetworkName": self.subnetwork_name or "",
            "filterBitsetNetworkAttributeName": self.filter_bitset_network_attribute_name
            or "",
            "traversabilityScope": _parse_enum(self.traversability_scope),
            "filterScope": _parse_enum(self.filter_scope),
            "conditionBarriers": self.condition_barriers or [],
            "functionBarriers": self.function_barriers or [],
            "filterBarriers": self.filter_barriers or [],
            "filterFunctionBarriers": self.filter_function_barriers or [],
            "functions": self.functions or [],
            "nearestNeighbor": self.nearest_neighbor or {},
            "outputFilters": self.output_filters or [],
            "outputConditions": self.output_conditions or [],
            "includeContainters": self.include_containers,
            "includeContent": self.include_content,
            "includeStructures": self.include_structures,
            "includeBarriers": self.include_barriers,
            "ValidateConsistency": self.validate_consistency,
            "validateLocatability": self.validate_locatability,
            "includeIsolated": self.include_isolated,
            "ignoreBarriersAtStartingPoints": self.ignore_barriers_at_starting_points,
            "includeUpToFirstSpatialContainer": self.include_up_to_first_spatial_container,
            "allowIndeterminateFlow": self.allow_indeterminate_flow,
        }

    def to_dict(self):
        return {
            "domainNetworkName": self.domain_network_name,
            "tierName": self.tier_name,
            "shortestPathNetworkAttributeName": self.shortest_path_network_attribute_name,
            "propagators": self.propagators,
            "targetTierName": self.target_tier_name,
            "subnetworkName": self.subnetwork_name,
            "filterBitsetNetworkAttributeName": self.filter_bitset_network_attribute_name,
            "traversabilityScope": _parse_enum(self.traversability_scope),
            "filterScope": _parse_enum(self.filter_scope),
            "conditionBarriers": self.condition_barriers,
            "functionBarriers": self.function_barriers,
            "filterBarriers": self.filter_barriers,
            "filterFunctionBarriers": self.filter_function_barriers,
            "functions": self.functions,
            "nearestNeighbor": self.nearest_neighbor,
            "outputFilters": self.output_filters,
            "outputConditions": self.output_conditions,
            "includeContainters": self.include_containers,
            "includeContent": self.include_content,
            "includeStructures": self.include_structures,
            "includeBarriers": self.include_barriers,
            "ValidateConsistency": self.validate_consistency,
            "validateLocatability": self.validate_locatability,
            "includeIsolated": self.include_isolated,
            "ignoreBarriersAtStartingPoints": self.ignore_barriers_at_starting_points,
            "includeUpToFirstSpatialContainer": self.include_up_to_first_spatial_container,
            "allowIndeterminateFlow": self.allow_indeterminate_flow,
        }

    @classmethod
    def from_config(cls, config) -> TraceConfiguration:
        """Loads the Trace Configurations from a Configuration"""
        return TraceConfiguration(
            domain_network_name=config["domainNetworkName"],
            tier_name=config["tierName"],
            shortest_path_network_attribute_name=config[
                "shortestPathNetworkAttributeName"
            ],
            propagators=config["propagators"],
            target_tier_name=config["targetTierName"],
            subnetwork_name=config["subnetworkName"],
            filter_bitset_network_attribute_name=config[
                "filterBitsetNetworkAttributeName"
            ],
            traversability_scope=config["traversabilityScope"],
            filter_scope=config["filterScope"],
            condition_barriers=config["conditionBarriers"],
            function_barriers=config["functionBarriers"],
            filter_barriers=config["filterBarriers"],
            filter_function_barriers=config["filterFunctionBarriers"],
            functions=config["functions"],
            nearest_neighbor=config["nearestNeighbor"],
            output_filters=config["outputFilters"],
            output_conditions=config["outputConditions"],
            include_containers=config["includeContainers"],
            include_content=config["includeContent"],
            include_structures=config["includeStructures"],
            include_barriers=config["includeBarriers"],
            validate_consistency=config["validateConsistency"],
            validate_locatability=config["validateLocatability"],
            include_isolated=config["includeIsolated"],
            ignore_barriers_at_starting_points=config["ignoreBarriersAtStartingPoints"],
            include_up_to_first_spatial_container=config[
                "includeUpToFirstSpatialContainer"
            ],
            allow_indeterminate_flow=config["allowIndeterminateFlow"],
        )



import logging
from .env import HAS_TENSORFLOW

if HAS_TENSORFLOW:
    import tensorflow as tf
    from tensorflow.keras.layers import (
        Input,
        Conv2D,
        Dropout,
        ReLU,
        BatchNormalization,
        UpSampling2D,
        Reshape,
        Layer,
    )
    from tensorflow.keras import Model
    from .._utils.fastai_tf_fit import (
        _tf_to_pytorch,
        _pytorch_to_tf_batch,
        _pytorch_to_tf,
    )
    from .common_tf import NormalizationLayerRGB, UpSample2DToSize

try:
    import torch
    import fastai
    import numpy as np
    from fastai.vision import imagenet_stats

    HAS_FASTAI = True
except:
    HAS_FASTAI = False


## Common start ##


## Common end ##

## Tensorflow specific utils start ##


def flatten_conv(output, n_anchors_per_cell, data_format):
    if data_format == "channels_last":
        x = output
    else:
        x = tf.transpose(output, perm=[0, 2, 3, 1])
    x = Reshape((-1, x.shape[-1] // n_anchors_per_cell))(x)
    return x


def get_ssd_head_output(ssd):
    data_format = (
        tf.keras.backend.image_data_format()
    )  #'channels_first' or 'channels_last'
    if data_format == "channels_last":
        bn_axis = -1
    else:
        bn_axis = 1

    n_anchors_per_cell = ssd._anchors_per_cell

    ## SSD Head

    # Zero Layer
    head_block0_drop1 = Dropout(0.3, name="head_block0_drop1")(
        ssd._backbone_initalized.output
    )

    # First Layer
    head_block1_conv1 = Conv2D(
        256,
        3,
        data_format=data_format,
        padding="same",
        activation="relu",
        name="head_block1_conv1",
    )(head_block0_drop1)
    head_block1_bn1 = BatchNormalization(axis=bn_axis, name="head_block1_bn1")(
        head_block1_conv1
    )
    head_block1_drop1 = Dropout(0.3, name="head_block1_drop1")(head_block1_bn1)

    outputs = []

    # Create Output Pipeline for each grid_size
    for grid in ssd.grids:
        # Grid upsampling Block
        head_block2_conv1 = Conv2D(
            256,
            3,
            data_format=data_format,
            padding="same",
            activation="relu",
            name=f"head_block2_conv1_{grid}",
        )(head_block1_drop1)
        head_block2_bn1 = BatchNormalization(
            axis=bn_axis, name=f"head_block2_bn1_{grid}"
        )(head_block2_conv1)
        head_block2_resize1 = UpSample2DToSize(
            grid, name=f"head_block2_resize1_{grid}"
        )(head_block2_bn1)
        head_block2_drop1 = Dropout(0.3, name=f"head_block2_drop1_{grid}")(
            head_block2_resize1
        )

        # Final Activations layer
        conv1 = Conv2D(
            ssd._data.c * n_anchors_per_cell,
            3,
            data_format=data_format,
            padding="same",
            name=f"head_block3_conv1_{grid}",
        )
        head_block3_conv1 = conv1(head_block2_drop1)
        conv2 = Conv2D(
            4 * n_anchors_per_cell,
            3,
            data_format=data_format,
            padding="same",
            name=f"head_block3_conv2_{grid}",
        )
        head_block3_conv2 = conv2(head_block2_drop1)

        # conv1.bias = conv1.bias - 4.

        head_block3_output1 = flatten_conv(
            head_block3_conv1, n_anchors_per_cell, data_format
        )
        head_block3_output2 = flatten_conv(
            head_block3_conv2, n_anchors_per_cell, data_format
        )

        outputs.append((head_block3_output1, head_block3_output2))

    output1 = tf.concat([a[0] for a in outputs], 1, name="concat1")
    output2 = tf.concat([a[1] for a in outputs], 1, name="concat2")
    return output1, output2


def remove_padded_y(image_y_bboxes, image_y_classes):
    try:
        bboxes = tf.reshape(image_y_bboxes, (-1, 4))
    except Exception:
        bboxes = tf.zeros((0, 4))
    # bboxes_to_keep = ( ( bboxes[:,2] - bboxes[:,0] ) != 0 ) tf 2.0.0
    bboxes_to_keep = tf.not_equal(bboxes[:, 2] - bboxes[:, 0], 0)
    return image_y_bboxes[bboxes_to_keep], image_y_classes[bboxes_to_keep]


def activations_to_bboxes(activations, anchors, grid_sizes):
    activation_bboxes = tf.tanh(activations)
    activation_centers = (activation_bboxes[:, :2] / 2 * grid_sizes) + anchors[:, :2]
    activation_height_widths = (activation_bboxes[:, 2:] / 2 + 1) * anchors[:, 2:]
    return height_width_to_corner(activation_centers, activation_height_widths)


def height_width_to_corner(centers, height_width):
    x = tf.concat([centers - (height_width / 2), centers + (height_width / 2)], axis=1)
    return x


def one_hot_embedding(labels, num_classes):
    return tf.gather(tf.eye(num_classes), labels)


def tf_loss_function_single_image(
    ssd, image_y_bboxes, image_y_classes, image_p_bboxes, image_p_classes
):
    image_y_bboxes, image_y_classes = remove_padded_y(image_y_bboxes, image_y_classes)

    # denormalize_bboxes
    image_y_bboxes = (image_y_bboxes + 1.0) / 2

    activation_bbox_corners = activations_to_bboxes(
        image_p_bboxes, ssd._anchors.cpu().numpy(), ssd._grid_sizes.cpu().numpy()
    )

    overlaps = ssd._jaccard(_tf_to_pytorch(image_y_bboxes), ssd._anchor_cnr.data)
    try:
        gt_overlap, gt_idx = ssd._map_to_ground_truth(overlaps, False)
        gt_idx = gt_idx.numpy()
    except Exception as e:
        return tf.constant(0.0), tf.constant(0.0)
    gt_clas = tf.gather(image_y_classes, gt_idx)
    pos = gt_overlap > 0.4
    pos_idx = torch.nonzero(pos)[:, 0].numpy()
    # gt_clas = tf.where( (pos.numpy()==0), 0, gt_clas) tf 2.0.0
    gt_clas = tf.compat.v2.where((pos.numpy() == 0), 0, gt_clas)
    ground_truth_classes = one_hot_embedding(gt_clas, ssd._data.c)
    gt_bbox = tf.gather(image_y_bboxes, gt_idx)

    location_loss = tf.reduce_sum(
        tf.abs(
            tf.gather(activation_bbox_corners, pos_idx) - tf.gather(gt_bbox, pos_idx)
        )
    )
    classification_loss = ssd._loss_function_classification(
        ground_truth_classes[:, 1:], image_p_classes[:, 1:]
    )
    return location_loss, classification_loss


class postProcessingLayer(Layer):
    def __init__(self, grid_sizes, anchors):
        super(postProcessingLayer, self).__init__()
        self.grid_sizes = grid_sizes
        self.anchors = anchors

    def __call__(self, preds):
        class_value_store = []
        confidence_store = []
        bbox_store = []
        batch_size = preds[0].shape[0]
        if batch_size is None:
            batch_size = 1
        for i in range(batch_size):
            class_values, confidence_scores, bboxes = TFOD_post_process_predictions(
                (preds[0][i], preds[1][i]), self.anchors, self.grid_sizes
            )
            class_value_store.append(class_values)
            confidence_store.append(confidence_scores)
            bbox_store.append(bboxes)
        return (
            tf.stack(class_value_store),
            tf.stack(confidence_store),
            tf.stack(bbox_store),
        )


def get_TFOD_post_processed_model(arcgis_model, input_normalization=True):
    model = arcgis_model.learn.model
    grid_sizes = _pytorch_to_tf(arcgis_model._grid_sizes)
    anchors = _pytorch_to_tf(arcgis_model._anchors)
    input_layer = model.input
    model_output = model.output

    if input_normalization:
        input_layer = Input(tuple(input_layer.shape[1:]))
        x = NormalizationLayerRGB()(input_layer)
        model_output = model(x)
    output_layer = postProcessingLayer(grid_sizes, anchors)(model_output)
    new_model = Model(input_layer, output_layer)
    return new_model


def TFOD_post_process_predictions(activations, anchors, grid_sizes):
    activation_classes, activation_bboxes = activations
    class_values = tf.argmax(activation_classes[:, 1:], axis=1)
    confidence_scores = tf.reduce_max(tf.sigmoid(activation_classes), axis=1)
    bboxes = TFOD_activations_to_bboxes(activation_bboxes, anchors, grid_sizes)
    return class_values, confidence_scores, bboxes


def TFOD_activations_to_bboxes(activation_bboxes, anchors, grid_sizes):
    activation_bboxes = tf.tanh(activation_bboxes)
    activation_centers = (activation_bboxes[:, :2] / 2 * grid_sizes) + anchors[:, :2]
    activation_hw = (activation_bboxes[:, 2:] / 2 + 1) * anchors[:, 2:]
    bboxes = tf.concat(
        [
            activation_centers - activation_hw / 2,
            activation_centers + activation_hw / 2,
        ],
        axis=1,
    )
    return bboxes


## Tensorflow specific utils end ##



from typing import Union, Dict, Any, Optional, ClassVar
from dataclasses import field, dataclass

from arcgis.realtime import Velocity
from arcgis.realtime.velocity.feeds._feed_template import _FeedTemplate
from arcgis.realtime.velocity.feeds.time import _HasTime, TimeInstant, TimeInterval


@dataclass
class StreamLayer(_FeedTemplate, _HasTime):
    """
    Receive features from a stream layer. This data class can be used to define the feed configuration and to
    create the feed.

    Data format is Esri stream layer. ArcGIS Velocity will automatically handle the location for you.

    ==================              ====================================================================
    **Parameter**                    **Description**
    ------------------              --------------------------------------------------------------------
    label                           String. Unique label for the feed instance.
    ------------------              --------------------------------------------------------------------
    description                     String. Feed description.
    ------------------              --------------------------------------------------------------------
    portal_item_id                  String. Portal item ID of the stream layer.
    ------------------              --------------------------------------------------------------------
    query                           String. Stream layer query parameters. The default is: "1=1"
    ------------------              --------------------------------------------------------------------
    fields                          String. Requested stream layer output fields.

                                    For example:

                                            "field1,field2"

                                    The default is: "*".
    ------------------              --------------------------------------------------------------------
    outSR                           int. Requested output spatial reference. The default is: 4326.

                                    .. note::
                                        To learn more about projected and geographic coordinate systems, refer to
                                        `Using spatial references <https://developers.arcgis.com/rest/services-reference/enterprise/using-spatial-references.htm>`_.
    ------------------              --------------------------------------------------------------------
    data_format                     String. Specifies the overall format of the incoming data.
    ==================              ====================================================================

    ==========================      ==================================================================================
    **Optional Argument**           **Description**
    ==========================      ==================================================================================
    WHERE clause                    String. Query to retrieve a subset of features.
    --------------------------      ----------------------------------------------------------------------------------
    Out fields                      String. Comma-separated list of fields to use for processing.
    --------------------------      ----------------------------------------------------------------------------------
    Output spatial reference        String. Spatial reference in which queried features should return.
    --------------------------      ----------------------------------------------------------------------------------
    extent                          Dict[str, Any]. JSON representing an envelope as defined by the ArcGIS
                                    REST API's JSON geometry schema.

                                    .. code-block:: python

                                        # Sample Value
                                        {
                                            "spatialReference": {
                                                "latestWkid": 3857,
                                                "wkid": 102100
                                            },
                                            "xmin": -14784278.027601289,
                                            "ymin": 2604610.848073723,
                                            "xmax": -11451317.846255329,
                                            "ymax": 6852675.132049575
                                        }

    --------------------------      ----------------------------------------------------------------------------------
    track_id_field                  String. Name of the field from the incoming data that should be set as
                                    track ID.
    --------------------------      ----------------------------------------------------------------------------------
    time                            [:class:`~arcgis.realtime.velocity.feeds.TimeInstant`, :class:`~arcgis.realtime.velocity.feeds.TimeInterval`]. An instance of time configuration that
                                    will be used to create time information from the incoming data.
    ==========================      ==================================================================================

    :return: A data class with stream layer feed configuration.

    .. code-block:: python

        # Usage Example

        from arcgis.realtime.velocity.feeds import StreamLayer
        from arcgis.realtime.velocity.feeds.time import TimeInterval, TimeInstant

        extent = {
            "spatialReference": {
                "latestWkid": 3857,
                "wkid": 102100
            },
            "xmin": "xmin",
            "ymin": "ymin",
            "xmax": "xmax",
            "ymax": "ymax"
        }

        stream_layer_config = StreamLayer(
            label="feed_name",
            description="feed_description",
            portal_item_id="portal_id",
            query="1=1",
            fields="*",
            outSR=4326,
            extent=extent
        )

        # use velocity object to get the FeedsManager instance
        feeds = velocity.feeds

        # use the FeedsManager object to create a feed from this feed configuration
        stream_layer_feed = feeds.create(stream_layer_config)
        stream_layer_feed.start()
        feeds.items

    """

    # fields that the user sets during init
    # Stream Layer specific properties
    portal_item_id: str
    query: str = field(default="1=1")
    fields: str = field(default="*")
    outSR: int = field(default=4326)
    extent: Optional[Dict[str, Any]] = None

    # FeedTemplate properties
    track_id_field: Optional[str] = None
    # HasTime properties
    time: Optional[Union[TimeInstant, TimeInterval]] = None
    # Stream Layer is a standard format and format properties do not need to be set in the feed configuration
    data_format: Any = field(default=None, init=False)

    # FeedTemplate properties
    _name: ClassVar[str] = "stream-layer"

    def __post_init__(self):
        if Velocity is None:
            return
        self._util = Velocity._util

        # validation of fields
        if self._util.is_valid(self.label) == False:
            raise ValueError(
                "Label should only contain alpha numeric, _ and space only"
            )

        # generate dictionary of this feed object's properties that will be used to query test-connection and
        # sample-messages Rest endpoint
        feed_properties = self._generate_feed_properties()

        test_connection = self._util.test_connection(
            input_type="feed", payload=feed_properties
        )
        # Test connection to make sure feed can fetch schema
        if test_connection is True:
            # test connection succeeded. Now try getting sample messages
            sample_payload = {
                "properties": {
                    "maxSamplesToCollect": 5,
                    "timeoutInMillis": 5000,
                }
            }
            self._dict_deep_merge(sample_payload, feed_properties)
            # Sample messages to fetch schema/fields
            sample_messages = self._util.sample_messages(
                input_type="feed", payload=sample_payload
            )

            if sample_messages["featureSchema"] is not None:
                # sample messages succeeded. Get Feature Schema from it
                self._set_fields(sample_messages["featureSchema"])

            print(
                "Feature Schema retrieved from the Feed:",
                sample_messages["featureSchema"],
            )

            # initiate actions for each of the following properties if it was set at init
            if self.track_id_field is not None:
                self.set_track_id(self.track_id_field)
            if self.time is not None:
                self.set_time_config(self.time)

    def _build(self) -> dict:
        feed_configuration = {
            "id": "",
            "label": self.label,
            "description": self.description,
            "feed": {**self._generate_schema_transformation()},
            "properties": {"executable": True},
        }

        feed_properties = self._generate_feed_properties()
        self._dict_deep_merge(feed_configuration["feed"], feed_properties)
        print(feed_configuration)
        return feed_configuration

    def _generate_feed_properties(self) -> dict:
        if self.extent:
            extent_properties = {f"{self._name}.extent": self.extent}
        else:
            extent_properties = {}

        feed_properties = {
            "name": self._name,
            "properties": {
                f"{self._name}.portalItemId": self.portal_item_id,
                f"{self._name}.query": self.query,
                f"{self._name}.fields": self.fields,
                f"{self._name}.outSR": self.outSR,
                **extent_properties,
            },
        }

        return feed_properties



"""
This resource lists all the portal machines in a site. Each portal machine
has a status that indicates whether the machine is ready to accept
requests.
"""
from ._base import BasePortalAdmin
from ..._impl.common._mixins import PropertyMap


########################################################################
class Machines(BasePortalAdmin):
    """
    This resource lists all the portal machines in a site. Each portal machine
    has a status that indicates whether the machine is ready to accept
    requests.
    """

    _pa = None
    _url = None
    _properties = None
    _json_dict = None
    _json = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis, portaladmin, **kwargs):
        """Constructor"""
        super(Machines, self).__init__(url=url, gis=gis, portaladmin=portaladmin)
        initialize = kwargs.pop("initialize", False)
        self._url = url
        self._pa = portaladmin
        if initialize:
            self._init()

    # ----------------------------------------------------------------------
    def list(self):
        """
        provides a list of all registered machines with the local GIS
        """
        machines = []
        for m in self.properties.machines:
            machines.append(
                Machine(
                    name=m["machineName"],
                    url=self._url,
                    info=dict(m),
                    gis=self._gis,
                    portaladmin=self._pa,
                )
            )
        return machines

    # ----------------------------------------------------------------------
    def get(self, name: str):
        """
        allows for retrieval of a single instance of Machine by it's
        registered name.
        """
        for m in self.properties.machines:
            if m["machineName"].lower() == name.lower():
                return Machine(
                    name=m["machineName"],
                    gis=self._gis,
                    url=self._url,
                    info=dict(m),
                    portaladmin=self._pa,
                )
        return


########################################################################
class Machine(BasePortalAdmin):
    """
    Represents a single machine instance registered with the GIS

    Parameters:
    :param name: machine name
    :param url: base url for the machine
    :param info: dictionary representing a single machine
    :param portaladmin: PortalAdminManager object
    :param initialize: (optional) if True, properties are loaded on
    creation
    """

    _pa = None
    _con = None
    _info = None
    _name = None
    _url = None
    _properties = None
    _json_dict = None
    _json = None

    # ----------------------------------------------------------------------
    def __init__(self, name, url, info, gis, portaladmin, **kwargs):
        """Constructor"""
        self._info = info
        super(Machine, self).__init__(
            url=url, gis=gis, name=name, info=info, portaladmin=portaladmin
        )
        self._info = info
        initialize = kwargs.pop("initialize", False)
        self._name = name
        self._url = url
        self._pa = portaladmin
        self._con = portaladmin._con
        self._gis = gis
        if initialize:
            self._init()

    # ----------------------------------------------------------------------
    def _init(self, connection=None):
        self._properties = PropertyMap(self._info)
        self._json_dict = self._info

    # ----------------------------------------------------------------------
    def status(self):
        """
        This operation checks whether a portal machine is ready to receive
        requests.
        """
        url = "%s/status/%s" % (self._url, self._name)
        params = {"f": "json"}
        res = self._con.get(path=url, params=params)
        if "status" in res:
            return res["status"] == "success"
        return False

    # ----------------------------------------------------------------------
    def unregister(self):
        """
        This operation unregisters a portal machine from a portal site. The
        operation can only performed when there are two machines
        participating in a portal site.
        """
        params = {"f": "json", "machineName": self._name}
        url = "%s/machines/unregister" % self._url
        return self._con.post(path=url, postdata=params)



try:
    import os, json
    import numpy as np
    import torch
    import torch.nn as nn
    import math
    from torch import tensor
    from .util import (
        variable_tile_size_check,
        fold_tensor,
        unfold_tensor,
        dihedral_transform,
        create_interpolation_mask,
    )

    HAS_TORCH = True
except Exception:
    HAS_TORCH = False

import arcgis
from arcgis.learn.models._multi_task_road_extractor import MultiTaskRoadExtractor

try:
    import arcpy
except Exception:
    pass


def split_tensor(tensor, tile_size, stride):
    # based on # https://discuss.pytorch.org/t/seemlessly-blending-tensors-together/65235/9
    mask = torch.ones_like(tensor)
    number_of_bands = tensor.shape[1]
    softmax_mask = torch.ones_like(tensor[0][0].unsqueeze(0).unsqueeze(0))

    unfold = nn.Unfold(kernel_size=(tile_size, tile_size), stride=stride)
    # Apply to mask and original image
    mask_p = unfold(softmax_mask)
    patches = unfold(tensor)

    patches = patches.reshape(number_of_bands, tile_size, tile_size, -1).permute(
        3, 0, 1, 2
    )
    if tensor.is_cuda:
        patches_base = torch.zeros(patches.size(), device=tensor.get_device())
    else:
        patches_base = torch.zeros(patches.size())

    return mask_p, patches_base, (tensor.size(2), tensor.size(3)), patches


def rebuild_tensor(input_tensor, mask_t, t_size, tile_size, stride):
    input_tensor_permuted = (
        input_tensor.permute(1, 2, 3, 0).reshape(-1, input_tensor.size(0)).unsqueeze(0)
    )
    fold = nn.Fold(
        output_size=(t_size[0], t_size[1]),
        kernel_size=(tile_size, tile_size),
        stride=stride,
    )
    output_tensor = fold(input_tensor_permuted) / fold(mask_t)
    return output_tensor


def calculate_rectangle_size_from_batch_size(batch_size):
    """
    calculate number of rows and cols to composite a rectangle given a batch size
    :param batch_size:
    :return: number of cols and number of rows
    """
    rectangle_height = int(math.sqrt(batch_size) + 0.5)
    rectangle_width = int(batch_size / rectangle_height)

    if rectangle_height * rectangle_width > batch_size:
        if rectangle_height >= rectangle_width:
            rectangle_height = rectangle_height - 1
        else:
            rectangle_width = rectangle_width - 1

    if (rectangle_height + 1) * rectangle_width <= batch_size:
        rectangle_height = rectangle_height + 1
    if (rectangle_width + 1) * rectangle_height <= batch_size:
        rectangle_width = rectangle_width + 1

    # swap col and row to make a horizontal rect
    if rectangle_height > rectangle_width:
        rectangle_height, rectangle_width = rectangle_width, rectangle_height

    if rectangle_height * rectangle_width != batch_size:
        return batch_size, 1

    return rectangle_height, rectangle_width


def get_tile_size(model_height, model_width, padding, batch_height, batch_width):
    """
    Calculate request tile size given model and batch dimensions
    :param model_height:
    :param model_width:
    :param padding:
    :param batch_width:
    :param batch_height:
    :return: tile height and tile width
    """
    tile_height = (model_height - 2 * padding) * batch_height
    tile_width = (model_width - 2 * padding) * batch_width

    return tile_height, tile_width


def tile_to_batch(
    pixel_block, model_height, model_width, padding, fixed_tile_size=True, **kwargs
):
    inner_width = model_width - 2 * padding
    inner_height = model_height - 2 * padding

    band_count, pb_height, pb_width = pixel_block.shape
    pixel_type = pixel_block.dtype

    if fixed_tile_size is True:
        batch_height = kwargs["batch_height"]
        batch_width = kwargs["batch_width"]
    else:
        batch_height = math.ceil((pb_height - 2 * padding) / inner_height)
        batch_width = math.ceil((pb_width - 2 * padding) / inner_width)

    batch = np.zeros(
        shape=(batch_width * batch_height, band_count, model_height, model_width),
        dtype=pixel_type,
    )
    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        # pixel block might not be the shape (band_count, model_height, model_width)
        sub_pixel_block = pixel_block[
            :,
            y * inner_height : y * inner_height + model_height,
            x * inner_width : x * inner_width + model_width,
        ]
        sub_pixel_block_shape = sub_pixel_block.shape
        batch[
            b, :, : sub_pixel_block_shape[1], : sub_pixel_block_shape[2]
        ] = sub_pixel_block

    return batch, batch_height, batch_width


def batch_to_tile(batch, batch_height, batch_width):
    batch_size, bands, inner_height, inner_width = batch.shape
    tile = np.zeros(
        shape=(bands, inner_height * batch_height, inner_width * batch_width),
        dtype=batch.dtype,
    )

    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        tile[
            :,
            y * inner_height : (y + 1) * inner_height,
            x * inner_width : (x + 1) * inner_width,
        ] = batch[b]

    return tile


class ChildImageClassifier:
    def initialize(self, model, model_as_file):
        if not HAS_TORCH:
            raise Exception(
                "Could not find the required deep learning dependencies. Ensure you have installed the required dependent libraries. See https://developers.arcgis.com/python/guide/deep-learning/"
            )

        if arcpy.env.processorType == "GPU" and torch.cuda.is_available():
            self.device = torch.device("cuda")
            arcgis.env._processorType = "GPU"
        else:
            self.device = torch.device("cpu")
            arcgis.env._processorType = "CPU"

        if model_as_file:
            with open(model, "r") as f:
                self.json_info = json.load(f)
        else:
            self.json_info = json.load(model)

        model_path = self.json_info["ModelFile"]
        if model_as_file and not os.path.isabs(model_path):
            model_path = os.path.abspath(
                os.path.join(os.path.dirname(model), model_path)
            )

        self.road_extractor = MultiTaskRoadExtractor.from_model(
            data=None, emd_path=model
        )
        self._learnmodel = self.road_extractor
        self.model = self.road_extractor.learn.model.to(self.device)
        self.model.eval()

    def getParameterInfo(self, required_parameters):
        required_parameters.extend(
            [
                {
                    "name": "padding",
                    "dataType": "numeric",
                    "value": int(self.json_info["ImageHeight"]) // 4,
                    "required": False,
                    "displayName": "Padding",
                    "description": "Padding",
                },
                {
                    "name": "batch_size",
                    "dataType": "numeric",
                    "required": False,
                    "value": 4,
                    "displayName": "Batch Size",
                    "description": "Batch Size",
                },
            ]
        )
        if (
            self.json_info.get("ArcGISLearnVersion", False)
            and self.json_info["ArcGISLearnVersion"] > "1.8.4"
        ):
            required_parameters.extend(
                [
                    {
                        "name": "return_probability_raster",
                        "dataType": "string",
                        "required": False,
                        "value": "False",
                        "displayName": "Return Probability Raster",
                        "description": "If True, will return the probability surface of the result.",
                    },
                    {
                        "name": "threshold",
                        "dataType": "numeric",
                        "value": 0.5,
                        "required": False,
                        "displayName": "Confidence Score Threshold [0.0, 1.0]",
                        "description": "Confidence score threshold value [0.0, 1.0]",
                    },
                ]
            )

        required_parameters.extend(
            [
                {
                    "name": "test_time_augmentation",
                    "dataType": "string",
                    "required": False,
                    "value": "False"
                    if "test_time_augmentation" not in self.json_info
                    else str(self.json_info["test_time_augmentation"]),
                    "displayName": "Perform test time augmentation while predicting",
                    "description": "If True, will merge predictions from flipped and rotated images.",
                },
                {
                    "name": "merge_policy",
                    "dataType": "string",
                    "required": False,
                    "value": "max",
                    "displayName": "Policy for merging augmented predictions",
                    "description": "Policy for merging predictions('mean', 'max' or 'min'). Applicable when test_time_augmentation is True.",
                },
            ]
        )

        required_parameters = variable_tile_size_check(
            self.json_info, required_parameters
        )
        return required_parameters

    def getConfiguration(self, **scalars):
        self.tytx = int(scalars.get("tile_size", self.json_info["ImageHeight"]))
        self.padding = int(
            scalars.get("padding", self.tytx // 4)
        )  ## Default padding Imageheight//4.
        self.batch_size = (
            int(math.sqrt(int(scalars.get("batch_size", 4)))) ** 2
        )  # Default 4 batch_size
        self.predict_background = scalars.get("predict_background", "true").lower() in [
            "true",
            "1",
            "t",
            "y",
            "yes",
        ]  # Default value True
        self.probability_raster = scalars.get(
            "return_probability_raster", "false"
        ).lower() in [
            "true",
            "1",
            "t",
            "y",
            "yes",
        ]  # Default value False
        (
            self.rectangle_height,
            self.rectangle_width,
        ) = calculate_rectangle_size_from_batch_size(self.batch_size)
        ty, tx = get_tile_size(
            self.tytx,
            self.tytx,
            self.padding,
            self.rectangle_height,
            self.rectangle_width,
        )

        self.thres = float(scalars.get("threshold", 0.5))  ## Default 0.5 threshold.

        self.use_tta = scalars.get("test_time_augmentation", "false").lower() in [
            "true",
            "1",
            "t",
            "y",
            "yes",
        ]  # Default value True

        self.merge_policy = scalars.get("merge_policy", "mean").lower()

        return {
            "extractBands": tuple(self.json_info["ExtractBands"]),
            "padding": self.padding,
            "tx": tx,
            "ty": ty,
            "threshold": self.thres,
            "fixedTileSize": 1,
            "test_time_augmentation": self.use_tta,
            "merge_policy": self.merge_policy,
        }

    def pixel_classify_image(
        self, model, tiles, device, classes, predict_bg, model_info
    ):
        model = model.to(device).eval()
        # logger.info("Tiles length is ", len(tiles))
        normed_batch_tensor = tensor(tiles).to(device).float()
        # logger.info("Normed length is ", len(normed_batch_tensor))
        with torch.no_grad():
            output, _ = model(normed_batch_tensor)

        if predict_bg:
            return output.max(dim=1)[1]
        else:
            output[:, 0] = -1
            return output.max(dim=1)[1]

    def updatePixels(self, tlc, shape, props, **pixelBlocks):  # 8 x 224 x 224 x 3
        input_image = pixelBlocks["raster_pixels"].astype(np.float32)
        batch, batch_height, batch_width = tile_to_batch(
            input_image,
            self.json_info["ImageHeight"],
            self.json_info["ImageWidth"],
            self.padding,
            fixed_tile_size=True,
            batch_height=self.rectangle_height,
            batch_width=self.rectangle_width,
        )

        semantic_predictions = self.pixel_classify_image(
            self.model,
            batch,
            self.device,
            classes=[clas["Name"] for clas in self.json_info["Classes"]],
            predict_bg=self.predict_background,
            model_info=self.json_info,
        )

        semantic_predictions = batch_to_tile(
            semantic_predictions.unsqueeze(dim=1).cpu().numpy(),
            batch_height,
            batch_width,
        )

        return semantic_predictions

    def split_predict_interpolate(self, normalized_image_tensor):
        model_arch = self.json_info["ModelParameters"]["mtl_model"]
        kernel_size = self.tytx
        stride = kernel_size - (2 * self.padding)

        if model_arch == "hourglass":
            kernel_size = (math.ceil(kernel_size / 2)) * 2

        # Split image into overlapping tiles
        masks, t_size, patches = unfold_tensor(
            normalized_image_tensor, kernel_size, stride
        )

        with torch.no_grad():
            output, _ = self.model(patches)

        interpolation_mask = create_interpolation_mask(
            kernel_size, 0, self.device, "hann"
        )
        output = output * interpolation_mask
        masks = masks * interpolation_mask

        # merge predictions from overlapping chips
        int_surface = fold_tensor(output, masks, t_size, kernel_size, stride)

        return int_surface

    def tta_predict(self, normalized_image_tensor, test_time_aug=True):
        all_activations = []

        transforms = [0]
        if test_time_aug:
            if self.json_info["ImageSpaceUsed"] == "MAP_SPACE":
                transforms = list(range(8))
            else:
                transforms = [
                    0,
                    2,
                ]  # no vertical flips for pixel space (oriented imagery)

        for k in transforms:
            flipped_image_tensor = dihedral_transform(normalized_image_tensor[0], k)
            int_surface = self.split_predict_interpolate(
                flipped_image_tensor.unsqueeze(0)
            )
            corrected_activation = dihedral_transform(int_surface[0], k)

            if k in [5, 6]:
                corrected_activation = dihedral_transform(int_surface[0], k).rot90(
                    2, [1, 2]
                )

            all_activations.append(corrected_activation)

        all_activations = torch.stack(all_activations)

        return all_activations

    def detectRoads(self, tlc, shape, props, **pixelBlocks):  # 8 x 224 x 224 x 3
        input_image = pixelBlocks["raster_pixels"].astype(np.float32)
        input_image_tensor = torch.tensor(input_image).to(self.device).float()

        normalized_image_tensor = input_image_tensor.unsqueeze(0)

        all_activations = self.tta_predict(
            normalized_image_tensor, test_time_aug=self.use_tta
        )

        softmax_surface = all_activations.softmax(dim=1)[:, [1], :, :]
        if self.merge_policy == "max":
            activations = softmax_surface.max(dim=0)[0]
        elif self.merge_policy == "min":
            activations = softmax_surface.min(dim=0)[0]
        else:
            activations = softmax_surface.mean(dim=0)

        if self.probability_raster:
            predictions = activations
        else:
            predictions = activations.gt(self.thres) * torch.tensor(1)

        pad = self.padding
        return (
            predictions[0][pad : -pad or None, pad : -pad or None]
            .unsqueeze(0)
            .cpu()
            .numpy()
        )



""" Defines store functions for working with assignments.
"""

from ... import workforce
from .dispatchers import query_dispatchers
from .utils import add_features, remove_features, update_features, validate


def get_assignment(project, object_id=None, global_id=None):
    """Gets the identified Assignment.  Exactly one form of identification should be provided.
    :param project:
    :param object_id: The assignment's OBJECTID.
    :param global_id: The assignment's GlobalID.
    :returns: Assignment
    """
    if object_id:
        where = "{} = {}".format(project._assignment_schema.object_id, object_id)
    elif global_id:
        where = "{} = '{}'".format(project._assignment_schema.global_id, global_id)
    else:
        where = "1=0"
    assignments = query_assignments(project, where)
    if assignments:
        return assignments[0]
    return None


def get_assignments(project):
    """Gets all Assignments in the project.
    :param project:
    :returns: list of Assignments
    """
    return query_assignments(project, "1=1")


def query_assignments(project, where="1=1"):
    """Executes a query against the assignments feature layer.
    :param project: The project in which to query assignments.
    :param where: An ArcGIS where clause.
    :returns: list of Assignments
    """
    assignments = []
    assignment_features = project.assignments_layer.query(
        where, return_all_records=True
    ).features
    # fetch assignment types, dispatchers, and workers
    project._update_cached_objects()
    # refresh cached objects
    for feature in assignment_features:
        assignments.append(workforce.Assignment(project, feature))
    return assignments


def add_assignments(project, assignments):
    """Adds Assignments to a project.

    Side effect: Upon successful addition on the server, the object_id and global_id fields of
    each Assignment in assignments will be updated to the values assigned by the server.

    :param project:
    :param assignments: list of Assignments
    :returns the list of Assignments
    :raises ValidationError: Indicates that one or more assignments failed validation.
    :raises ServerError: Indicates that the server rejected the assignments.
    """
    project._update_cached_objects()
    use_global_ids = True
    for assignment in assignments:
        validate(assignment._validate_for_add)
        if assignment.global_id is None:
            use_global_ids = False

    features = [assignment.feature for assignment in assignments]
    add_features(project.assignments_layer, features, use_global_ids)
    return assignments


def add_assignment(
    project,
    feature=None,
    geometry=None,
    assignment_type=None,
    assigned_date=None,
    assignment_read=None,
    completed_date=None,
    declined_comment=None,
    declined_date=None,
    description=None,
    dispatcher=None,
    due_date=None,
    in_progress_date=None,
    location=None,
    notes=None,
    paused_date=None,
    priority=None,
    status=None,
    work_order_id=None,
    worker=None,
):
    """
    Adds a new assignment to the project
    """
    project._update_cached_objects()
    assignment = workforce.Assignment(
        project,
        feature,
        geometry,
        assignment_type,
        assigned_date,
        assignment_read,
        completed_date,
        declined_comment,
        declined_date,
        description,
        dispatcher,
        due_date,
        in_progress_date,
        location,
        notes,
        paused_date,
        priority,
        status,
        work_order_id,
        worker,
    )

    return add_assignments(project, [assignment])[0]


def update_assignments(project, assignments):
    """Updates Assignments.
    :param project:
    :param assignments: list of Assignments to update
    :raises ValidationError: Indicates that one or more assignments failed validation.
    :raises ServerError: Indicates that the server rejected the updates.
    """
    project._update_cached_objects()
    for assignment in assignments:
        validate(assignment._validate_for_update)
    features = [assignment.feature for assignment in assignments]
    update_features(project.assignments_layer, features)
    return assignments


def update_assignment(
    project,
    assignment,
    geometry=None,
    assignment_type=None,
    assigned_date=None,
    assignment_read=None,
    completed_date=None,
    declined_comment=None,
    declined_date=None,
    description=None,
    dispatcher=None,
    due_date=None,
    in_progress_date=None,
    location=None,
    notes=None,
    paused_date=None,
    priority=None,
    status=None,
    work_order_id=None,
    worker=None,
):
    """
    Sets the properties of an assignment and updates the item on the server
    """
    project._update_cached_objects()
    if status:
        # logic for resetting stale dates if reassigning
        if assignment.status == "declined" and worker and worker != assignment.worker:
            assignment.declined_date = None
        if assignment.status == "paused" and worker and worker != assignment.worker:
            assignment.paused_date = None
        assignment.status = status
    if geometry:
        assignment.geometry = geometry
    if assigned_date:
        assignment.assigned_date = assigned_date
    if assignment_read:
        assignment.assignment_read = assignment_read
    if completed_date:
        assignment.completed_date = completed_date
    if declined_comment:
        assignment.declined_comment = declined_comment
    if declined_date:
        assignment.declined_date = declined_date
    if description:
        assignment.description = description
    if due_date:
        assignment.due_date = due_date
    if in_progress_date:
        assignment.in_progress_date = in_progress_date
    if location:
        assignment.location = location
    if notes:
        assignment.notes = notes
    if paused_date:
        assignment.paused_date = paused_date
    if priority:
        assignment.priority = priority
    if status:
        assignment.status = status
    if work_order_id:
        assignment.work_order_id = work_order_id
    if assignment_type:
        assignment.assignment_type = assignment_type
    if dispatcher:
        assignment.dispatcher = dispatcher
    if worker:
        assignment.worker = worker
    return update_assignments(project, [assignment])[0]


def delete_assignments(project, assignments):
    """Removes assignments from the project.
    :param project:
    :param assignments: list of Assignments
    :raises ValidationError: Indicates that one or more assignments failed validation.
    :raises ServerError: Indicates that the server rejected the removal.
    """
    project._update_cached_objects()
    for assignment in assignments:
        validate(assignment._validate_for_remove)
    features = [assignment.feature for assignment in assignments]
    remove_features(project.assignments_layer, features)


def dispatchers_for_assignment_features(project, features):
    dispatcher_ids = [
        feature.attributes.get(project._assignment_schema.dispatcher_id)
        for feature in features
    ]
    dispatcher_ids = [did for did in dispatcher_ids if did is not None]
    if dispatcher_ids:
        dispatchers_where = "{} IN ({})".format(
            project._dispatcher_schema.object_id, ",".join(map(str, dispatcher_ids))
        )
        return query_dispatchers(project, dispatchers_where)
    return []


def workers_for_assignment_features(project, features):
    worker_ids = [
        feature.attributes.get(project._assignment_schema.worker_id)
        for feature in features
    ]
    worker_ids = [wid for wid in worker_ids if wid is not None]
    if worker_ids:
        workers_where = "{} IN ({})".format(
            project._worker_schema.object_id, ",".join(map(str, worker_ids))
        )
        return workforce._store.query_workers(project, workers_where)
    return []



"""Backport of python 3.8 functools.cached_property.

cached_property() - computed once per instance, cached as attribute
"""

__all__ = ("cached_property",)

# Standard Library
from sys import version_info

if version_info >= (3, 8):
    # Standard Library
    from functools import cached_property  # pylint: disable=no-name-in-module
else:
    # Standard Library
    from threading import RLock
    from typing import Any
    from typing import Callable
    from typing import Optional
    from typing import Type
    from typing import TypeVar

    _NOT_FOUND = object()
    _T = TypeVar("_T")
    _S = TypeVar("_S")

    # noinspection PyPep8Naming
    class cached_property:  # NOSONAR  # pylint: disable=invalid-name  # noqa: N801
        """Cached property implementation.

        Transform a method of a class into a property whose value is computed once
        and then cached as a normal attribute for the life of the instance.
        Similar to property(), with the addition of caching.
        Useful for expensive computed properties of instances
        that are otherwise effectively immutable.
        """

        def __init__(self, func: Callable[[Any], _T]) -> None:
            """Cached property implementation."""
            self.func = func
            self.attrname: Optional[str] = None
            self.__doc__ = func.__doc__
            self.lock = RLock()

        def __set_name__(self, owner: Type[Any], name: str) -> None:
            """Assign attribute name and owner."""
            if self.attrname is None:
                self.attrname = name
            elif name != self.attrname:
                raise TypeError(
                    "Cannot assign the same cached_property to two different names "
                    f"({self.attrname!r} and {name!r})."
                )

        def __get__(self, instance, owner=None) -> Any:
            """Property-like getter implementation.

            :return: property instance if requested on class or value/cached value if requested on instance.

            :rtype: Union[cached_property[_T], _T]
            :raises TypeError: call without calling __set_name__ or no '__dict__' attribute
            """
            if instance is None:
                return self
            if self.attrname is None:
                raise TypeError(
                    "Cannot use cached_property instance without calling __set_name__ on it."
                )
            try:
                cache = instance.__dict__
            except (
                AttributeError
            ):  # not all objects have __dict__ (e.g. class defines slots)
                msg = (
                    f"No '__dict__' attribute on {type(instance).__name__!r} "
                    f"instance to cache {self.attrname!r} property."
                )
                raise TypeError(msg) from None
            val = cache.get(self.attrname, _NOT_FOUND)
            if val is _NOT_FOUND:
                with self.lock:
                    # check if another thread filled cache while we awaited lock
                    val = cache.get(self.attrname, _NOT_FOUND)
                    if val is _NOT_FOUND:
                        val = self.func(instance)
                        try:
                            cache[self.attrname] = val
                        except TypeError:
                            msg = (
                                f"The '__dict__' attribute on {type(instance).__name__!r} instance "
                                f"does not support item assignment for caching {self.attrname!r} property."
                            )
                            raise TypeError(msg) from None
            return val



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# tridentnet_r50_caffe_mstrain_3x_coco.py, box AP=40.3

_base_ = "./_base_/models/faster_rcnn_r50_caffe_c4.py"

model = dict(
    type="TridentFasterRCNN",
    pretrained="open-mmlab://detectron2/resnet50_caffe",
    backbone=dict(
        type="TridentResNet",
        trident_dilations=(1, 2, 3),
        num_branch=3,
        test_branch_idx=1,
    ),
    roi_head=dict(type="TridentRoIHead", num_branch=3, test_branch_idx=1),
    train_cfg=dict(
        rpn_proposal=dict(nms_post=500, max_num=500),
        rcnn=dict(sampler=dict(num=128, pos_fraction=0.5, add_gt_as_proposals=False)),
    ),
)

checkpoint = "https://download.openmmlab.com/mmdetection/v2.0/tridentnet/tridentnet_r50_caffe_mstrain_3x_coco/tridentnet_r50_caffe_mstrain_3x_coco_20201130_100539-46d227ba.pth"



# Referred from https://github.com/timeseriesAI/timeseriesAI
import torch
import torch.nn as nn
from .layers import *


class Block(nn.Module):
    def __init__(self, ni, nf, ks=[7, 5, 3], act_fn="relu"):
        super().__init__()
        self.conv1 = convlayer(ni, nf, ks[0], act_fn=act_fn)
        self.conv2 = convlayer(nf, nf, ks[1], act_fn=act_fn)
        self.conv3 = convlayer(nf, nf, ks[2], act_fn=False)

        # expand channels for the sum if necessary
        self.shortcut = noop if ni == nf else convlayer(ni, nf, ks=1, act_fn=False)
        self.act_fn = get_act_layer(act_fn)

    def forward(self, x):
        res = x
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        sc = self.shortcut(res)
        x += sc
        x = self.act_fn(x)
        return x


class _TSResCNN(nn.Module):
    def __init__(self, c_in, c_out):
        nf = 64
        super().__init__()
        self.block = Block(c_in, nf, ks=[7, 5, 3], act_fn="relu")
        self.conv1 = convlayer(nf, nf * 2, ks=3, act_fn="leakyrelu", negative_slope=0.2)
        self.conv2 = convlayer(nf * 2, nf * 4, ks=3, act_fn="prelu")
        self.conv3 = convlayer(nf * 4, nf * 2, ks=3, act_fn="elu", alpha=0.3)
        self.gap = nn.AdaptiveAvgPool1d(1)
        self.lin = nn.Linear(nf * 2, c_out)

    def forward(self, x):
        x = self.block(x)
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.gap(x).squeeze(-1)
        return self.lin(x)
